@@ -1,6 +1,6 @@
 
 /*
- * $Id: forward.c,v 1.82.2.1 2002/04/01 12:51:27 hno Exp $
+ * $Id: forward.c,v 1.131.2.2 2008/01/23 11:39:16 hno Exp $
  *
  * DEBUG: section 17    Request Forwarding
  * AUTHOR: Duane Wessels
@@ -36,12 +36,20 @@
 
 #include "squid.h"
 
+#if LINUX_NETFILTER
+#include <linux/netfilter_ipv4.h>
+#endif
+#if LINUX_TPROXY
+#include <linux/netfilter_ipv4/ip_tproxy.h>
+#endif
+
 static PSC fwdStartComplete;
 static void fwdDispatch(FwdState *);
 static void fwdConnectStart(void *);	/* should be same as EVH */
 static void fwdStateFree(FwdState * fwdState);
 static PF fwdConnectTimeout;
 static PF fwdServerClosed;
+static PF fwdPeerClosed;
 static CNCB fwdConnectDone;
 static int fwdCheckRetry(FwdState * fwdState);
 static int fwdReforward(FwdState *);
@@ -98,10 +106,12 @@ fwdStateFree(FwdState * fwdState)
 	    fwdState->err = NULL;
 	} else {
 	    EBIT_CLR(e->flags, ENTRY_FWD_HDR_WAIT);
-	    storeComplete(e);
 	    storeReleaseRequest(e);
+	    storeComplete(e);
 	}
     }
+    if (EBIT_TEST(e->flags, ENTRY_DEFER_READ))
+	storeResetDefer(e);
     if (storePendingNClients(e) > 0)
 	assert(!EBIT_TEST(e->flags, ENTRY_FWD_HDR_WAIT));
     p = fwdStateServerPeer(fwdState);
@@ -119,8 +129,6 @@ fwdStateFree(FwdState * fwdState)
 	fwdState->server_fd = -1;
 	debug(17, 3) ("fwdStateFree: closing FD %d\n", sfd);
 	comm_close(sfd);
-	if (p)
-	    p->stats.conn_open--;
     }
     cbdataFree(fwdState);
 }
@@ -128,13 +136,17 @@ fwdStateFree(FwdState * fwdState)
 static int
 fwdCheckRetry(FwdState * fwdState)
 {
+    if (shutting_down)
+	return 0;
     if (fwdState->entry->store_status != STORE_PENDING)
 	return 0;
     if (fwdState->entry->mem_obj->inmem_hi > 0)
 	return 0;
     if (fwdState->n_tries > 10)
 	return 0;
-    if (squid_curtime - fwdState->start > Config.Timeout.connect)
+    if (fwdState->origin_tries > 2)
+	return 0;
+    if (squid_curtime - fwdState->start >= Config.Timeout.forward)
 	return 0;
     if (fwdState->flags.dont_retry)
 	return 0;
@@ -143,6 +155,33 @@ fwdCheckRetry(FwdState * fwdState)
     return 1;
 }
 
+static int
+fwdCheckRetriable(FwdState * fwdState)
+{
+    /* If there is a request body then Squid can only try once
+     * even if the method is indempotent
+     */
+    if (fwdState->request->body_reader)
+	return 0;
+
+    /* RFC2616 9.1 Safe and Idempotent Methods */
+    switch (fwdState->request->method) {
+	/* 9.1.1 Safe Methods */
+    case METHOD_GET:
+    case METHOD_HEAD:
+	/* 9.1.2 Indepontent Methods */
+    case METHOD_PUT:
+    case METHOD_DELETE:
+    case METHOD_OPTIONS:
+    case METHOD_TRACE:
+	break;
+    default:
+	return 0;
+    }
+
+    return 1;
+}
+
 static void
 fwdServerClosed(int fd, void *data)
 {
@@ -150,7 +189,10 @@ fwdServerClosed(int fd, void *data)
     debug(17, 2) ("fwdServerClosed: FD %d %s\n", fd, storeUrl(fwdState->entry));
     assert(fwdState->server_fd == fd);
     fwdState->server_fd = -1;
+    if (EBIT_TEST(fwdState->entry->flags, ENTRY_DEFER_READ))
+	storeResetDefer(fwdState->entry);
     if (fwdCheckRetry(fwdState)) {
+	int originserver = (fwdState->servers->peer == NULL);
 	debug(17, 3) ("fwdServerClosed: re-forwarding (%d tries, %d secs)\n",
 	    fwdState->n_tries,
 	    (int) (squid_curtime - fwdState->start));
@@ -168,26 +210,130 @@ fwdServerClosed(int fd, void *data)
 		/* Use next. The last "direct" entry is retried multiple times */
 		fwdState->servers = fs->next;
 		fwdServerFree(fs);
+		originserver = 0;
 	    }
 	}
-	/* use eventAdd to break potential call sequence loops */
-	eventAdd("fwdConnectStart", fwdConnectStart, fwdState, 0.0, 0);
-    } else {
-	fwdStateFree(fwdState);
+	/* use eventAdd to break potential call sequence loops and to slow things down a little */
+	eventAdd("fwdConnectStart", fwdConnectStart, fwdState, originserver ? 0.05 : 0.005, 0);
+	return;
     }
+    if (!fwdState->err && shutting_down)
+	fwdState->err = errorCon(ERR_SHUTTING_DOWN, HTTP_GATEWAY_TIMEOUT, fwdState->request);
+    fwdStateFree(fwdState);
 }
 
+#if USE_SSL
+static void
+fwdNegotiateSSL(int fd, void *data)
+{
+    FwdState *fwdState = data;
+    FwdServer *fs = fwdState->servers;
+    SSL *ssl = fd_table[fd].ssl;
+    int ret;
+    ErrorState *err;
+    request_t *request = fwdState->request;
+
+    errno = 0;
+    ERR_clear_error();
+    if ((ret = SSL_connect(ssl)) <= 0) {
+	int ssl_error = SSL_get_error(ssl, ret);
+	switch (ssl_error) {
+	case SSL_ERROR_WANT_READ:
+	    commSetSelect(fd, COMM_SELECT_READ, fwdNegotiateSSL, fwdState, 0);
+	    return;
+	case SSL_ERROR_WANT_WRITE:
+	    commSetSelect(fd, COMM_SELECT_WRITE, fwdNegotiateSSL, fwdState, 0);
+	    return;
+	default:
+	    debug(81, 1) ("fwdNegotiateSSL: Error negotiating SSL connection on FD %d: %s (%d/%d/%d)\n", fd, ERR_error_string(ERR_get_error(), NULL), ssl_error, ret, errno);
+	    err = errorCon(ERR_CONNECT_FAIL, HTTP_BAD_GATEWAY, request);
+#ifdef EPROTO
+	    err->xerrno = EPROTO;
+#else
+	    err->xerrno = EACCES;
+#endif
+	    fwdFail(fwdState, err);
+	    if (fs->peer) {
+		peerConnectFailed(fs->peer);
+		fs->peer->stats.conn_open--;
+	    }
+	    comm_close(fd);
+	    return;
+	}
+    }
+    if (fs->peer && !SSL_session_reused(ssl)) {
+	if (fs->peer->sslSession)
+	    SSL_SESSION_free(fs->peer->sslSession);
+	fs->peer->sslSession = SSL_get1_session(ssl);
+    }
+#if NOT_YET
+    if (verify_domain) {
+	char *host;
+	STACK_OF(GENERAL_NAME) * altnames;
+	if (fs->peer) {
+	    if (fs->peer->ssldomain)
+		host = fs->peer->ssldomain;
+	    else
+		host = fs->peer->host;
+	} else {
+	    host = fs->request->host;
+	}
+	if (!ssl_verify_domain(host, ssl)) {
+	    debug(17, 1) ("Warning: SSL certificate does not match host name '%s'\n", host);
+	}
+    }
+#endif
+    fwdDispatch(fwdState);
+}
+
+static void
+fwdInitiateSSL(FwdState * fwdState)
+{
+    FwdServer *fs = fwdState->servers;
+    int fd = fwdState->server_fd;
+    SSL *ssl;
+    SSL_CTX *sslContext = NULL;
+    peer *peer = fs->peer;
+    if (peer) {
+	assert(peer->use_ssl);
+	sslContext = peer->sslContext;
+    } else {
+	sslContext = Config.ssl_client.sslContext;
+    }
+    assert(sslContext);
+    if ((ssl = SSL_new(sslContext)) == NULL) {
+	ErrorState *err;
+	debug(83, 1) ("fwdInitiateSSL: Error allocating handle: %s\n",
+	    ERR_error_string(ERR_get_error(), NULL));
+	err = errorCon(ERR_SOCKET_FAILURE, HTTP_INTERNAL_SERVER_ERROR, fwdState->request);
+	err->xerrno = errno;
+	fwdFail(fwdState, err);
+	fwdStateFree(fwdState);
+	return;
+    }
+    SSL_set_fd(ssl, fd);
+    if (peer) {
+	if (peer->sslSession)
+	    SSL_set_session(ssl, peer->sslSession);
+    }
+    fd_table[fd].ssl = ssl;
+    fd_table[fd].read_method = &ssl_read_method;
+    fd_table[fd].write_method = &ssl_write_method;
+    fd_note(fd, "Negotiating SSL");
+    fwdNegotiateSSL(fd, fwdState);
+}
+#endif
+
 static void
 fwdConnectDone(int server_fd, int status, void *data)
 {
     FwdState *fwdState = data;
-    static FwdState *current = NULL;
     FwdServer *fs = fwdState->servers;
     ErrorState *err;
     request_t *request = fwdState->request;
-    assert(current != fwdState);
-    current = fwdState;
     assert(fwdState->server_fd == server_fd);
+    if (Config.onoff.log_ip_on_direct && status != COMM_ERR_DNS && fs->code == HIER_DIRECT)
+	hierarchyNote(&fwdState->request->hier, fs->code, fd_table[server_fd].ipaddr);
     if (status == COMM_ERR_DNS) {
 	/*
 	 * Only set the dont_retry flag if the DNS lookup fails on
@@ -198,46 +344,29 @@ fwdConnectDone(int server_fd, int status, void *data)
 	    fwdState->flags.dont_retry = 1;
 	debug(17, 4) ("fwdConnectDone: Unknown host: %s\n",
 	    request->host);
-	err = errorCon(ERR_DNS_FAIL, HTTP_SERVICE_UNAVAILABLE);
+	err = errorCon(ERR_DNS_FAIL, HTTP_GATEWAY_TIMEOUT, fwdState->request);
 	err->dnsserver_msg = xstrdup(dns_error_message);
-	err->request = requestLink(request);
 	fwdFail(fwdState, err);
-	if (fs->peer)
-	    fs->peer->stats.conn_open--;
 	comm_close(server_fd);
     } else if (status != COMM_OK) {
 	assert(fs);
-	err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE);
+	err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT, fwdState->request);
 	err->xerrno = errno;
-	if (fs->peer) {
-	    err->host = xstrdup(fs->peer->host);
-	    err->port = fs->peer->http_port;
-	} else {
-	    err->host = xstrdup(request->host);
-	    err->port = request->port;
-	}
-	err->request = requestLink(request);
 	fwdFail(fwdState, err);
-	if (fs->peer) {
+	if (fs->peer)
 	    peerConnectFailed(fs->peer);
-	    fs->peer->stats.conn_open--;
-	}
 	comm_close(server_fd);
     } else {
 	debug(17, 3) ("fwdConnectDone: FD %d: '%s'\n", server_fd, storeUrl(fwdState->entry));
-	if (fs->peer)
-	    hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->host);
-	else if (Config.onoff.log_ip_on_direct)
-	    hierarchyNote(&fwdState->request->hier, fs->code, fd_table[server_fd].ipaddr);
-	else
-	    hierarchyNote(&fwdState->request->hier, fs->code, request->host);
-	fd_note(server_fd, storeUrl(fwdState->entry));
-	fd_table[server_fd].uses++;
-	if (fs->peer)
-	    peerConnectSucceded(fs->peer);
+#if USE_SSL
+	if ((fs->peer && fs->peer->use_ssl) ||
+	    (!fs->peer && request->protocol == PROTO_HTTPS)) {
+	    fwdInitiateSSL(fwdState);
+	    return;
+	}
+#endif
 	fwdDispatch(fwdState);
     }
-    current = NULL;
 }
 
 static void
@@ -245,13 +374,14 @@ fwdConnectTimeout(int fd, void *data)
 {
     FwdState *fwdState = data;
     StoreEntry *entry = fwdState->entry;
+    FwdServer *fs = fwdState->servers;
     ErrorState *err;
-    peer *p = fwdStateServerPeer(fwdState);
     debug(17, 2) ("fwdConnectTimeout: FD %d: '%s'\n", fd, storeUrl(entry));
     assert(fd == fwdState->server_fd);
+    if (Config.onoff.log_ip_on_direct && fs->code == HIER_DIRECT && fd_table[fd].ipaddr[0])
+	hierarchyNote(&fwdState->request->hier, fs->code, fd_table[fd].ipaddr);
     if (entry->mem_obj->inmem_hi == 0) {
-	err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT);
-	err->request = requestLink(fwdState->request);
+	err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT, fwdState->request);
 	err->xerrno = ETIMEDOUT;
 	fwdFail(fwdState, err);
 	/*
@@ -261,16 +391,85 @@ fwdConnectTimeout(int fd, void *data)
 	    if (fwdState->servers->peer)
 		peerConnectFailed(fwdState->servers->peer);
     }
-    if (p)
-	p->stats.conn_open--;
     comm_close(fd);
 }
 
+typedef struct {
+    peer *peer;
+    char *domain;
+} idleconn;
+
+static void
+free_idleconn(void *data)
+{
+    idleconn *idle = data;
+    safe_free(idle->domain)
+}
+
+CBDATA_TYPE(idleconn);
+
+static void
+fwdConnectIdleDone(int fd, int status, void *data)
+{
+    idleconn *idle = data;
+    idle->peer->stats.idle_opening--;
+    if (fd >= 0)
+	pconnPush(fd, idle->peer->name, idle->peer->http_port, idle->domain, NULL, 0);
+    cbdataFree(idle);
+}
+
+static void
+fwdConnectIdleTimeout(int fd, void *data)
+{
+    idleconn *idle = data;
+    debug(17, 2) ("fwdConnectIdleTimeout: FD %d: '%s'\n", fd, idle->peer->name);
+    idle->peer->stats.idle_opening--;
+    comm_close(fd);
+    cbdataFree(idle);
+}
+
+static void
+openIdleConn(peer * peer, const char *domain, struct in_addr outgoing, unsigned short tos, int ctimeout)
+{
+    int fd = comm_openex(SOCK_STREAM,
+	IPPROTO_TCP,
+	outgoing,
+	0,
+	COMM_NONBLOCKING,
+	tos,
+	peer->name);
+    idleconn *idle;
+    if (fd < 0) {
+	debug(50, 4) ("openIdleConn: %s\n", xstrerror());
+	return;;
+    }
+    CBDATA_INIT_TYPE_FREECB(idleconn, free_idleconn);
+    idle = cbdataAlloc(idleconn);
+    idle->peer = peer;
+    cbdataLock(idle->peer);
+    idle->domain = xstrdup(domain);
+    /*
+     * stats.conn_open is used to account for the number of
+     * connections that we have open to the peer, so we can limit
+     * based on the max-conn option.  We need to increment here,
+     * even if the connection may fail.
+     */
+    peer->stats.conn_open++;
+    peer->stats.idle_opening++;
+    comm_add_close_handler(fd, fwdPeerClosed, peer);
+    commSetTimeout(fd,
+	ctimeout,
+	fwdConnectIdleTimeout,
+	idle);
+    commConnectStart(fd, peer->host, peer->http_port, fwdConnectIdleDone, idle);
+}
+
 static struct in_addr
 aclMapAddr(acl_address * head, aclCheck_t * ch)
 {
     acl_address *l;
     struct in_addr addr;
+    aclChecklistCacheInit(ch);
     for (l = head; l; l = l->next) {
 	if (aclMatchAclList(l->acl_list, ch))
 	    return l->addr;
@@ -283,6 +482,7 @@ static int
 aclMapTOS(acl_tos * head, aclCheck_t * ch)
 {
     acl_tos *l;
+    aclChecklistCacheInit(ch);
     for (l = head; l; l = l->next) {
 	if (aclMatchAclList(l->acl_list, ch))
 	    return l->tos;
@@ -296,9 +496,6 @@ getOutgoingAddr(request_t * request)
     aclCheck_t ch;
     memset(&ch, '\0', sizeof(aclCheck_t));
     if (request) {
-	ch.src_addr = request->client_addr;
-	ch.my_addr = request->my_addr;
-	ch.my_port = request->my_port;
 	ch.request = request;
     }
     return aclMapAddr(Config.accessList.outgoing_address, &ch);
@@ -310,9 +507,6 @@ getOutgoingTOS(request_t * request)
     aclCheck_t ch;
     memset(&ch, '\0', sizeof(aclCheck_t));
     if (request) {
-	ch.src_addr = request->client_addr;
-	ch.my_addr = request->my_addr;
-	ch.my_port = request->my_port;
 	ch.request = request;
     }
     return aclMapTOS(Config.accessList.outgoing_tos, &ch);
@@ -323,50 +517,126 @@ fwdConnectStart(void *data)
 {
     FwdState *fwdState = data;
     const char *url = storeUrl(fwdState->entry);
-    int fd;
+    int fd = -1;
     ErrorState *err;
     FwdServer *fs = fwdState->servers;
     const char *host;
+    const char *name;
     unsigned short port;
-    time_t ctimeout;
+    const char *domain = NULL;
+    int ctimeout;
+    int ftimeout = Config.Timeout.forward - (squid_curtime - fwdState->start);
     struct in_addr outgoing;
     unsigned short tos;
+#if LINUX_TPROXY
+    struct in_tproxy itp;
+#endif
+    int idle = -1;
+
     assert(fs);
     assert(fwdState->server_fd == -1);
     debug(17, 3) ("fwdConnectStart: %s\n", url);
     if (fs->peer) {
 	host = fs->peer->host;
+	name = fs->peer->name;
 	port = fs->peer->http_port;
+	if (fs->peer->options.originserver)
+	    domain = fwdState->request->host;
+	else
+	    domain = "*";
 	ctimeout = fs->peer->connect_timeout > 0 ? fs->peer->connect_timeout
 	    : Config.Timeout.peer_connect;
-    } else if (fwdState->request->flags.accelerated &&
-	Config.Accel.single_host && Config.Accel.host) {
-	host = Config.Accel.host;
-	port = Config.Accel.port;
-	ctimeout = Config.Timeout.connect;
     } else {
-	host = fwdState->request->host;
+	host = name = fwdState->request->host;
 	port = fwdState->request->port;
 	ctimeout = Config.Timeout.connect;
     }
-    if ((fd = pconnPop(host, port)) >= 0) {
-	debug(17, 3) ("fwdConnectStart: reusing pconn FD %d\n", fd);
-	fwdState->server_fd = fd;
-	fwdState->n_tries++;
-	comm_add_close_handler(fd, fwdServerClosed, fwdState);
-	fwdConnectDone(fd, COMM_OK, fwdState);
+    if (ftimeout < 0)
+	ftimeout = 5;
+    if (ftimeout < ctimeout)
+	ctimeout = ftimeout;
+    fwdState->request->flags.pinned = 0;
+    if (fs->code == PINNED) {
+	int auth;
+	fd = clientGetPinnedConnection(fwdState->request->pinned_connection, fwdState->request, fs->peer, &auth);
+	if (fd >= 0) {
+#if 0
+	    if (!fs->peer)
+		fs->code = HIER_DIRECT;
+#endif
+	    fwdState->server_fd = fd;
+	    fwdState->n_tries++;
+	    fwdState->request->flags.pinned = 1;
+	    if (auth)
+		fwdState->request->flags.auth = 1;
+	    comm_add_close_handler(fd, fwdServerClosed, fwdState);
+	    fwdConnectDone(fd, COMM_OK, fwdState);
+	    return;
+	}
+	/* Failure. Fall back on next path */
+	cbdataUnlock(fwdState->request->pinned_connection);
+	fwdState->request->pinned_connection = NULL;
+	fwdState->servers = fs->next;
+	fwdServerFree(fs);
+	fwdConnectStart(fwdState);
 	return;
     }
+#if LINUX_TPROXY
+    if (fd == -1 && fwdState->request->flags.tproxy)
+	fd = pconnPop(name, port, domain, &fwdState->request->client_addr, 0, NULL);
+#endif
+    if (fd == -1) {
+	fd = pconnPop(name, port, domain, NULL, 0, &idle);
+    }
+    if (fd != -1) {
+	if (fwdCheckRetriable(fwdState)) {
+	    debug(17, 3) ("fwdConnectStart: reusing pconn FD %d\n", fd);
+	    fwdState->server_fd = fd;
+	    fwdState->n_tries++;
+	    if (!fs->peer)
+		fwdState->origin_tries++;
+	    comm_add_close_handler(fd, fwdServerClosed, fwdState);
+	    if (fs->peer)
+		hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->name);
+	    else if (Config.onoff.log_ip_on_direct && fs->code == HIER_DIRECT)
+		hierarchyNote(&fwdState->request->hier, fs->code, fd_table[fd].ipaddr);
+	    else
+		hierarchyNote(&fwdState->request->hier, fs->code, name);
+	    if (fs->peer && idle >= 0 && idle < fs->peer->idle) {
+		debug(17, 3) ("fwdConnectStart: Opening idle connetions for '%s'\n",
+		    fs->peer->name);
+		outgoing = getOutgoingAddr(fwdState->request);
+		tos = getOutgoingTOS(fwdState->request);
+		debug(17, 3) ("fwdConnectStart: got addr %s, tos %d\n",
+		    inet_ntoa(outgoing), tos);
+		idle += fs->peer->stats.idle_opening;
+		while (idle < fs->peer->idle) {
+		    openIdleConn(fs->peer, domain, outgoing, tos, ctimeout);
+		    idle++;
+		}
+	    }
+	    fwdDispatch(fwdState);
+	    return;
+	} else {
+	    /* Discard the persistent connection to not cause
+	     * a imbalance in number of conenctions open if there
+	     * is a lot of POST requests
+	     */
+	    comm_close(fd);
+	}
+    }
 #if URL_CHECKSUM_DEBUG
     assert(fwdState->entry->mem_obj->chksum == url_checksum(url));
 #endif
     outgoing = getOutgoingAddr(fwdState->request);
     tos = getOutgoingTOS(fwdState->request);
 
+    fwdState->request->out_ip = outgoing;
+
     debug(17, 3) ("fwdConnectStart: got addr %s, tos %d\n",
 	inet_ntoa(outgoing), tos);
     fd = comm_openex(SOCK_STREAM,
-	0,
+	IPPROTO_TCP,
 	outgoing,
 	0,
 	COMM_NONBLOCKING,
@@ -374,28 +644,62 @@ fwdConnectStart(void *data)
 	url);
     if (fd < 0) {
 	debug(50, 4) ("fwdConnectStart: %s\n", xstrerror());
-	err = errorCon(ERR_SOCKET_FAILURE, HTTP_INTERNAL_SERVER_ERROR);
+	err = errorCon(ERR_SOCKET_FAILURE, HTTP_INTERNAL_SERVER_ERROR, fwdState->request);
 	err->xerrno = errno;
-	err->request = requestLink(fwdState->request);
 	fwdFail(fwdState, err);
 	fwdStateFree(fwdState);
 	return;
     }
     fwdState->server_fd = fd;
     fwdState->n_tries++;
+    if (!fs->peer)
+	fwdState->origin_tries++;
     /*
      * stats.conn_open is used to account for the number of
      * connections that we have open to the peer, so we can limit
      * based on the max-conn option.  We need to increment here,
      * even if the connection may fail.
      */
-    if (fs->peer)
+    if (fs->peer) {
 	fs->peer->stats.conn_open++;
+	comm_add_close_handler(fd, fwdPeerClosed, fs->peer);
+    }
     comm_add_close_handler(fd, fwdServerClosed, fwdState);
     commSetTimeout(fd,
 	ctimeout,
 	fwdConnectTimeout,
 	fwdState);
+    if (fs->peer) {
+	hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->name);
+    } else {
+#if LINUX_TPROXY
+	if (fwdState->request->flags.tproxy) {
+
+	    itp.v.addr.faddr.s_addr = fwdState->src.sin_addr.s_addr;
+	    itp.v.addr.fport = 0;
+
+	    /* If these syscalls fail then we just fallback to connecting
+	     * normally by simply ignoring the errors...
+	     */
+	    itp.op = TPROXY_ASSIGN;
+	    if (setsockopt(fd, SOL_IP, IP_TPROXY, &itp, sizeof(itp)) == -1) {
+		debug(20, 1) ("tproxy ip=%s,0x%x,port=%d ERROR ASSIGN\n",
+		    inet_ntoa(itp.v.addr.faddr),
+		    itp.v.addr.faddr.s_addr,
+		    itp.v.addr.fport);
+	    } else {
+		itp.op = TPROXY_FLAGS;
+		itp.v.flags = ITP_CONNECT;
+		if (setsockopt(fd, SOL_IP, IP_TPROXY, &itp, sizeof(itp)) == -1) {
+		    debug(20, 1) ("tproxy ip=%x,port=%d ERROR CONNECT\n",
+			itp.v.addr.faddr.s_addr,
+			itp.v.addr.fport);
+		}
+	    }
+	}
+#endif
+	hierarchyNote(&fwdState->request->hier, fs->code, fwdState->request->host);
+    }
     commConnectStart(fd, host, port, fwdConnectDone, fwdState);
 }
 
@@ -417,8 +721,7 @@ fwdStartFail(FwdState * fwdState)
 {
     ErrorState *err;
     debug(17, 3) ("fwdStartFail: %s\n", storeUrl(fwdState->entry));
-    err = errorCon(ERR_CANNOT_FORWARD, HTTP_SERVICE_UNAVAILABLE);
-    err->request = requestLink(fwdState->request);
+    err = errorCon(ERR_CANNOT_FORWARD, HTTP_GATEWAY_TIMEOUT, fwdState->request);
     err->xerrno = errno;
     fwdFail(fwdState, err);
     fwdStateFree(fwdState);
@@ -431,28 +734,43 @@ fwdDispatch(FwdState * fwdState)
     request_t *request = fwdState->request;
     StoreEntry *entry = fwdState->entry;
     ErrorState *err;
+    int server_fd = fwdState->server_fd;
+    FwdServer *fs = fwdState->servers;
     debug(17, 3) ("fwdDispatch: FD %d: Fetching '%s %s'\n",
 	fwdState->client_fd,
-	RequestMethodStr[request->method],
+	RequestMethods[request->method].str,
 	storeUrl(entry));
-    /*assert(!EBIT_TEST(entry->flags, ENTRY_DISPATCHED)); */
-    assert(entry->ping_status != PING_WAITING);
-    assert(entry->lock_count);
-    EBIT_SET(entry->flags, ENTRY_DISPATCHED);
-    netdbPingSite(request->host);
     /*
      * Assert that server_fd is set.  This is to guarantee that fwdState
      * is attached to something and will be deallocated when server_fd
      * is closed.
      */
-    assert(fwdState->server_fd > -1);
+    assert(server_fd > -1);
+    /*assert(!EBIT_TEST(entry->flags, ENTRY_DISPATCHED)); */
+    assert(entry->ping_status != PING_WAITING);
+    assert(entry->lock_count);
+    EBIT_SET(entry->flags, ENTRY_DISPATCHED);
+    fd_note(server_fd, storeUrl(fwdState->entry));
+    fd_table[server_fd].uses++;
+    if (fd_table[server_fd].uses == 1 && fs->peer)
+	peerConnectSucceded(fs->peer);
+    fwdState->request->out_ip = fd_table[server_fd].local_addr;
+    netdbPingSite(request->host);
+    entry->mem_obj->refresh_timestamp = squid_curtime;
     if (fwdState->servers && (p = fwdState->servers->peer)) {
 	p->stats.fetches++;
 	fwdState->request->peer_login = p->login;
+	fwdState->request->peer_domain = p->domain;
 	httpStart(fwdState);
     } else {
 	fwdState->request->peer_login = NULL;
+	fwdState->request->peer_domain = NULL;
 	switch (request->protocol) {
+#if USE_SSL
+	case PROTO_HTTPS:
+	    httpStart(fwdState);
+	    break;
+#endif
 	case PROTO_HTTP:
 	    httpStart(fwdState);
 	    break;
@@ -462,9 +780,6 @@ fwdDispatch(FwdState * fwdState)
 	case PROTO_FTP:
 	    ftpStart(fwdState);
 	    break;
-	case PROTO_WAIS:
-	    waisStart(fwdState);
-	    break;
 	case PROTO_CACHEOBJ:
 	case PROTO_INTERNAL:
 	case PROTO_URN:
@@ -473,11 +788,11 @@ fwdDispatch(FwdState * fwdState)
 	case PROTO_WHOIS:
 	    whoisStart(fwdState);
 	    break;
+	case PROTO_WAIS:	/* not implemented */
 	default:
 	    debug(17, 1) ("fwdDispatch: Cannot retrieve '%s'\n",
 		storeUrl(entry));
-	    err = errorCon(ERR_UNSUP_REQ, HTTP_BAD_REQUEST);
-	    err->request = requestLink(request);
+	    err = errorCon(ERR_UNSUP_REQ, HTTP_BAD_REQUEST, fwdState->request);
 	    fwdFail(fwdState, err);
 	    /*
 	     * Force a persistent connection to be closed because
@@ -490,11 +805,6 @@ fwdDispatch(FwdState * fwdState)
 	     * transient (network) error; its a bug.
 	     */
 	    fwdState->flags.dont_retry = 1;
-	    /*
-	     * this assertion exists because if we are connected to
-	     * a peer, then we need to decrement p->stats.conn_open.
-	     */
-	    assert(NULL == p);
 	    comm_close(fwdState->server_fd);
 	    break;
 	}
@@ -517,8 +827,14 @@ fwdReforward(FwdState * fwdState)
 	debug(17, 3) ("fwdReforward: No, ENTRY_FWD_HDR_WAIT isn't set\n");
 	return 0;
     }
+    if (fwdState->request->flags.pinned) {
+	debug(17, 3) ("fwdReforward: No, Connection oriented authentication is used\n");
+	return 0;
+    }
     if (fwdState->n_tries > 9)
 	return 0;
+    if (fwdState->origin_tries > 1)
+	return 0;
     if (fwdState->request->flags.body_sent)
 	return 0;
     assert(fs);
@@ -545,11 +861,35 @@ fwdServersFree(FwdServer ** FS)
     }
 }
 
+void
+fwdStartPeer(peer * p, StoreEntry * e, request_t * r)
+{
+    FwdState *fwdState;
+    FwdServer *peer = NULL;
+    debug(17, 3) ("fwdStartPeer: '%s'\n", storeUrl(e));
+    e->mem_obj->request = requestLink(r);
+#if URL_CHECKSUM_DEBUG
+    assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
+#endif
+    fwdState = cbdataAlloc(FwdState);
+    fwdState->entry = e;
+    fwdState->client_fd = -1;
+    fwdState->server_fd = -1;
+    fwdState->request = requestLink(r);
+    fwdState->start = squid_curtime;
+    fwdState->orig_entry_flags = e->flags;
+    storeLockObject(e);
+    if (!fwdState->request->flags.pinned)
+	EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
+    storeRegisterAbort(e, fwdAbort, fwdState);
+    peerAddFwdServer(&peer, p, HIER_DIRECT);
+    fwdStartComplete(peer, fwdState);
+}
+
 void
 fwdStart(int fd, StoreEntry * e, request_t * r)
 {
     FwdState *fwdState;
-    aclCheck_t ch;
     int answer;
     ErrorState *err;
     /*
@@ -557,34 +897,30 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
      * from peer_digest.c, asn.c, netdb.c, etc and should always
      * be allowed.  yuck, I know.
      */
-    if (r->client_addr.s_addr != no_addr.s_addr) {
+    if (r->client_addr.s_addr != no_addr.s_addr && r->protocol != PROTO_INTERNAL && r->protocol != PROTO_CACHEOBJ) {
 	/*      
 	 * Check if this host is allowed to fetch MISSES from us (miss_access)
 	 */
-	memset(&ch, '\0', sizeof(aclCheck_t));
-	ch.src_addr = r->client_addr;
-	ch.my_addr = r->my_addr;
-	ch.my_port = r->my_port;
-	ch.request = r;
-	answer = aclCheckFast(Config.accessList.miss, &ch);
+	answer = aclCheckFastRequest(Config.accessList.miss, r);
 	if (answer == 0) {
-	    err = errorCon(ERR_FORWARDING_DENIED, HTTP_FORBIDDEN);
-	    err->request = requestLink(r);
-	    err->src_addr = r->client_addr;
+	    err_type page_id;
+	    page_id = aclGetDenyInfoPage(&Config.denyInfoList, AclMatchedName, 1);
+	    if (page_id == ERR_NONE)
+		page_id = ERR_FORWARDING_DENIED;
+	    err = errorCon(page_id, HTTP_FORBIDDEN, r);
 	    errorAppendEntry(e, err);
 	    return;
 	}
     }
     debug(17, 3) ("fwdStart: '%s'\n", storeUrl(e));
-    e->mem_obj->request = requestLink(r);
-    e->mem_obj->fd = fd;
+    if (!e->mem_obj->request)
+	e->mem_obj->request = requestLink(r);
 #if URL_CHECKSUM_DEBUG
     assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
 #endif
     if (shutting_down) {
 	/* more yuck */
-	err = errorCon(ERR_SHUTTING_DOWN, HTTP_SERVICE_UNAVAILABLE);
-	err->request = requestLink(r);
+	err = errorCon(ERR_SHUTTING_DOWN, HTTP_GATEWAY_TIMEOUT, r);
 	errorAppendEntry(e, err);
 	return;
     }
@@ -610,8 +946,19 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
     fwdState->server_fd = -1;
     fwdState->request = requestLink(r);
     fwdState->start = squid_curtime;
+    fwdState->orig_entry_flags = e->flags;
+
+#if LINUX_TPROXY
+    /* If we need to transparently proxy the request
+     * then we need the client source address and port */
+    fwdState->src.sin_family = AF_INET;
+    fwdState->src.sin_addr = r->client_addr;
+    fwdState->src.sin_port = r->client_port;
+#endif
+
     storeLockObject(e);
-    EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
+    if (!fwdState->request->flags.pinned)
+	EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
     storeRegisterAbort(e, fwdAbort, fwdState);
     peerSelect(r, e, fwdStartComplete, fwdState);
 }
@@ -634,8 +981,14 @@ fwdCheckDeferRead(int fd, void *data)
 	(void) 0;
     else {
 	int i = delayMostBytesWanted(mem, INT_MAX);
-	if (0 == i)
+	if (0 == i) {
+	    /* No storeDeferRead here as it's not store dependent.
+	     * Will get changed the day delay pools interact nicely
+	     * with the store..
+	     */
+	    commDeferFD(fd);
 	    return 1;
+	}
 	/* was: rc = -(rc != INT_MAX); */
 	else if (INT_MAX == i)
 	    rc = 0;
@@ -643,17 +996,33 @@ fwdCheckDeferRead(int fd, void *data)
 	    rc = -1;
     }
 #endif
+    if (EBIT_TEST(e->flags, ENTRY_DEFER_READ)) {
+	commDeferFD(mem->serverfd);
+	return 1;
+    }
     if (EBIT_TEST(e->flags, ENTRY_FWD_HDR_WAIT))
 	return rc;
-    if (mem->inmem_hi - storeLowestMemReaderOffset(e) < READ_AHEAD_GAP)
-	return rc;
-    return 1;
+    if (EBIT_TEST(e->flags, RELEASE_REQUEST)) {
+	/* Just a small safety cap to defer storing more data into the object
+	 * if there already is way too much. This handles the case when there
+	 * is disk clients pending on a too large object being fetched and a
+	 * few other corner cases.
+	 */
+	if (fd >= 0 && mem->inmem_hi - mem->inmem_lo > SM_PAGE_SIZE + Config.Store.maxInMemObjSize + Config.readAheadGap) {
+	    storeDeferRead(e, fd);
+	    return 1;
+	}
+    }
+    if (fd >= 0 && mem->inmem_hi - storeLowestMemReaderOffset(e) > Config.readAheadGap) {
+	storeDeferRead(e, fd);
+	return 1;
+    }
+    return rc;
 }
 
 void
 fwdFail(FwdState * fwdState, ErrorState * errorState)
 {
-    assert(EBIT_TEST(fwdState->entry->flags, ENTRY_FWD_HDR_WAIT));
     debug(17, 3) ("fwdFail: %s \"%s\"\n\t%s\n",
 	err_type_str[errorState->type],
 	httpStatusString(errorState->http_status),
@@ -661,6 +1030,8 @@ fwdFail(FwdState * fwdState, ErrorState * errorState)
     if (fwdState->err)
 	errorStateFree(fwdState->err);
     fwdState->err = errorState;
+    if (!errorState->request)
+	errorState->request = requestLink(fwdState->request);
 }
 
 /*
@@ -674,6 +1045,16 @@ fwdAbort(void *data)
     fwdStateFree(fwdState);
 }
 
+/*
+ * Accounts for closed persistent connections
+ */
+static void
+fwdPeerClosed(int fd, void *data)
+{
+    peer *p = data;
+    p->stats.conn_open--;
+}
+
 /*
  * Frees fwdState without closing FD or generating an abort
  */
@@ -683,6 +1064,9 @@ fwdUnregister(int fd, FwdState * fwdState)
     debug(17, 3) ("fwdUnregister: %s\n", storeUrl(fwdState->entry));
     assert(fd == fwdState->server_fd);
     assert(fd > -1);
+    commSetDefer(fd, NULL, NULL);
+    if (EBIT_TEST(fwdState->entry->flags, ENTRY_DEFER_READ))
+	storeResumeRead(fwdState->entry);
     comm_remove_close_handler(fd, fwdServerClosed, fwdState);
     fwdState->server_fd = -1;
 }
@@ -711,6 +1095,24 @@ fwdComplete(FwdState * fwdState)
 	if (fwdState->server_fd > -1)
 	    fwdUnregister(fwdState->server_fd, fwdState);
 	storeEntryReset(e);
+	/* HACK WARNING: This fiddling with the flags really
+	 * should be done in the store layer, but current
+	 * design does not allow it to be done proper in a
+	 * sane manner.
+	 * A sign that we have pushed the design of everything
+	 * going via a single StoreEntry per request a bit too far.
+	 */
+	if (EBIT_TEST(e->flags, ENTRY_NEGCACHED)) {
+	    storeSetPrivateKey(e);
+	    EBIT_CLR(e->flags, ENTRY_NEGCACHED);
+	    EBIT_CLR(e->flags, KEY_EARLY_PUBLIC);
+	}
+	if (EBIT_TEST(e->flags, RELEASE_REQUEST)) {
+	    EBIT_CLR(e->flags, RELEASE_REQUEST);
+	    EBIT_CLR(e->flags, KEY_EARLY_PUBLIC);
+	}
+	if (EBIT_TEST(fwdState->orig_entry_flags, ENTRY_CACHABLE))
+	    EBIT_SET(e->flags, ENTRY_CACHABLE);
 	fwdStartComplete(fwdState->servers, fwdState);
     } else {
 	debug(17, 3) ("fwdComplete: not re-forwarding status %d\n",
@@ -779,13 +1181,14 @@ int
 fwdReforwardableStatus(http_status s)
 {
     switch (s) {
+    case HTTP_BAD_GATEWAY:
+    case HTTP_GATEWAY_TIMEOUT:
+	return 1;
     case HTTP_FORBIDDEN:
     case HTTP_INTERNAL_SERVER_ERROR:
     case HTTP_NOT_IMPLEMENTED:
-    case HTTP_BAD_GATEWAY:
     case HTTP_SERVICE_UNAVAILABLE:
-    case HTTP_GATEWAY_TIMEOUT:
-	return 1;
+	return Config.retry.onerror;
     default:
 	return 0;
     }
@@ -818,7 +1221,7 @@ fwdLog(FwdState * fwdState)
 	(int) current_time.tv_sec,
 	(int) current_time.tv_usec / 1000,
 	fwdState->last_status,
-	RequestMethodStr[fwdState->request->method],
+	RequestMethods[fwdState->request->method].str,
 	fwdState->request->canonical);
 }
 
