@@ -1,6 +1,6 @@
 
 /*
- * $Id: store_dir_aufs.c,v 1.40.2.6 2002/08/23 22:31:53 hno Exp $
+ * $Id: store_dir_aufs.c,v 1.70.2.1 2008/01/02 15:49:32 hno Exp $
  *
  * DEBUG: section 47    Store Directory Routines
  * AUTHOR: Duane Wessels
@@ -35,6 +35,7 @@
 
 #include "squid.h"
 
+#include "async_io.h"
 #include "store_asyncufs.h"
 
 #define DefaultLevelOneDirs     16
@@ -50,7 +51,6 @@ struct _RebuildState {
     int curlvl1;
     int curlvl2;
     struct {
-	unsigned int need_to_validate:1;
 	unsigned int clean:1;
 	unsigned int init:1;
     } flags;
@@ -64,7 +64,7 @@ struct _RebuildState {
     struct _store_rebuild_data counts;
 };
 
-static int n_asyncufs_dirs = 0;
+int n_asyncufs_dirs = 0;
 static int *asyncufs_dir_index = NULL;
 MemPool *squidaio_state_pool = NULL;
 MemPool *aufs_qread_pool = NULL;
@@ -82,7 +82,7 @@ static EVH storeAufsDirRebuildFromSwapLog;
 static int storeAufsDirGetNextFile(RebuildState *, sfileno *, int *size);
 static StoreEntry *storeAufsDirAddDiskRestore(SwapDir * SD, const cache_key * key,
     sfileno file_number,
-    size_t swap_file_sz,
+    squid_file_sz swap_file_sz,
     time_t expires,
     time_t timestamp,
     time_t lastref,
@@ -106,12 +106,12 @@ static STNEWFS storeAufsDirNewfs;
 static STDUMP storeAufsDirDump;
 static STMAINTAINFS storeAufsDirMaintain;
 static STCHECKOBJ storeAufsDirCheckObj;
+static STCHECKLOADAV storeAufsDirCheckLoadAv;
 static STREFOBJ storeAufsDirRefObj;
 static STUNREFOBJ storeAufsDirUnrefObj;
 static QS rev_int_sort;
 static int storeAufsDirClean(int swap_index);
 static EVH storeAufsDirCleanEvent;
-static int storeAufsDirIs(SwapDir * sd);
 static int storeAufsFilenoBelongsHere(int fn, int F0, int F1, int F2);
 static int storeAufsCleanupDoubleCheck(SwapDir *, StoreEntry *);
 static void storeAufsDirStats(SwapDir *, StoreEntry *);
@@ -215,7 +215,11 @@ storeAufsDirCreateDirectory(const char *path, int should_exist)
 	} else {
 	    fatalf("Swap directory %s is not a directory.", path);
 	}
+#ifdef _SQUID_MSWIN_
+    } else if (0 == mkdir(path)) {
+#else
     } else if (0 == mkdir(path, 0755)) {
+#endif
 	debug(47, should_exist ? 1 : 3) ("%s created\n", path);
 	created = 1;
     } else {
@@ -320,6 +324,10 @@ storeAufsDirOpenSwapLog(SwapDir * sd)
     char *path;
     int fd;
     path = storeAufsDirSwapLogFile(sd, NULL);
+    if (aioinfo->swaplog_fd >= 0) {
+	debug(50, 1) ("storeAufsDirOpenSwapLog: %s already open\n", path);
+	return;
+    }
     fd = file_open(path, O_WRONLY | O_CREAT | O_BINARY);
     if (fd < 0) {
 	debug(50, 1) ("%s: %s\n", path, xstrerror());
@@ -327,10 +335,6 @@ storeAufsDirOpenSwapLog(SwapDir * sd)
     }
     debug(50, 3) ("Cache Dir #%d log opened on FD %d\n", sd->index, fd);
     aioinfo->swaplog_fd = fd;
-    if (0 == n_asyncufs_dirs)
-	assert(NULL == asyncufs_dir_index);
-    n_asyncufs_dirs++;
-    assert(n_asyncufs_dirs <= Config.cacheSwap.n_configured);
 }
 
 static void
@@ -343,12 +347,16 @@ storeAufsDirCloseSwapLog(SwapDir * sd)
     debug(47, 3) ("Cache Dir #%d log closed on FD %d\n",
 	sd->index, aioinfo->swaplog_fd);
     aioinfo->swaplog_fd = -1;
-    n_asyncufs_dirs--;
-    assert(n_asyncufs_dirs >= 0);
-    if (0 == n_asyncufs_dirs)
-	safe_free(asyncufs_dir_index);
 }
 
+static void
+storeAufsCheckConfig(SwapDir * sd)
+{
+    if (!opt_create_swap_dirs)
+	requirePathnameExists("cache_dir", sd->path);
+}
+
+
 static void
 storeAufsDirInit(SwapDir * sd)
 {
@@ -360,6 +368,7 @@ storeAufsDirInit(SwapDir * sd)
     storeAufsDirInitBitmap(sd);
     if (storeAufsDirVerifyCacheDirs(sd) < 0)
 	fatal(errmsg);
+    squidaio_init();
     storeAufsDirOpenSwapLog(sd);
     storeAufsDirRebuild(sd);
     if (!started_clean_event) {
@@ -369,6 +378,24 @@ storeAufsDirInit(SwapDir * sd)
     (void) storeDirGetBlkSize(sd->path, &sd->fs.blksize);
 }
 
+static void
+storeAufsDirRebuildComplete(RebuildState * rb)
+{
+    if (rb->log) {
+	debug(47, 1) ("Done reading %s swaplog (%d entries)\n",
+	    rb->sd->path, rb->n_read);
+	fclose(rb->log);
+	rb->log = NULL;
+    } else {
+	debug(47, 1) ("Done scanning %s (%d entries)\n",
+	    rb->sd->path, rb->counts.scancount);
+    }
+    store_dirs_rebuilding--;
+    storeAufsDirCloseTmpSwapLog(rb->sd);
+    storeRebuildComplete(&rb->counts);
+    cbdataFree(rb);
+}
+
 static void
 storeAufsDirRebuildFromDirectory(void *data)
 {
@@ -377,7 +404,7 @@ storeAufsDirRebuildFromDirectory(void *data)
     LOCAL_ARRAY(char, hdr_buf, SM_PAGE_SIZE);
     StoreEntry *e = NULL;
     StoreEntry tmpe;
-    cache_key key[MD5_DIGEST_CHARS];
+    cache_key key[SQUID_MD5_DIGEST_LENGTH];
     sfileno filn = 0;
     int count;
     int size;
@@ -392,12 +419,7 @@ storeAufsDirRebuildFromDirectory(void *data)
 	assert(fd == -1);
 	fd = storeAufsDirGetNextFile(rb, &filn, &size);
 	if (fd == -2) {
-	    debug(47, 1) ("Done scanning %s swaplog (%d entries)\n",
-		rb->sd->path, rb->n_read);
-	    store_dirs_rebuilding--;
-	    storeAufsDirCloseTmpSwapLog(rb->sd);
-	    storeRebuildComplete(&rb->counts);
-	    cbdataFree(rb);
+	    storeAufsDirRebuildComplete(rb);
 	    return;
 	} else if (fd < 0) {
 	    continue;
@@ -441,18 +463,47 @@ storeAufsDirRebuildFromDirectory(void *data)
 	    continue;
 	}
 	debug(47, 3) ("storeAufsDirRebuildFromDirectory: successful swap meta unpacking\n");
-	memset(key, '\0', MD5_DIGEST_CHARS);
+	memset(key, '\0', SQUID_MD5_DIGEST_LENGTH);
 	memset(&tmpe, '\0', sizeof(StoreEntry));
 	for (t = tlv_list; t; t = t->next) {
 	    switch (t->type) {
 	    case STORE_META_KEY:
-		assert(t->length == MD5_DIGEST_CHARS);
-		xmemcpy(key, t->value, MD5_DIGEST_CHARS);
+		assert(t->length == SQUID_MD5_DIGEST_LENGTH);
+		xmemcpy(key, t->value, SQUID_MD5_DIGEST_LENGTH);
 		break;
+#if SIZEOF_SQUID_FILE_SZ == SIZEOF_SIZE_T
 	    case STORE_META_STD:
 		assert(t->length == STORE_HDR_METASIZE);
 		xmemcpy(&tmpe.timestamp, t->value, STORE_HDR_METASIZE);
 		break;
+#else
+	    case STORE_META_STD_LFS:
+		assert(t->length == STORE_HDR_METASIZE);
+		xmemcpy(&tmpe.timestamp, t->value, STORE_HDR_METASIZE);
+		break;
+	    case STORE_META_STD:
+		assert(t->length == STORE_HDR_METASIZE_OLD);
+		{
+		    struct {
+			time_t timestamp;
+			time_t lastref;
+			time_t expires;
+			time_t lastmod;
+			size_t swap_file_sz;
+			u_short refcount;
+			u_short flags;
+		    }     *tmp = t->value;
+		    assert(sizeof(*tmp) == STORE_HDR_METASIZE_OLD);
+		    tmpe.timestamp = tmp->timestamp;
+		    tmpe.lastref = tmp->lastref;
+		    tmpe.expires = tmp->expires;
+		    tmpe.lastmod = tmp->lastmod;
+		    tmpe.swap_file_sz = tmp->swap_file_sz;
+		    tmpe.refcount = tmp->refcount;
+		    tmpe.flags = tmp->flags;
+		}
+		break;
+#endif
 	    default:
 		break;
 	    }
@@ -526,21 +577,10 @@ storeAufsDirRebuildFromSwapLog(void *data)
     /* load a number of objects per invocation */
     for (count = 0; count < rb->speed; count++) {
 	if (fread(&s, ss, 1, rb->log) != 1) {
-	    debug(47, 1) ("Done reading %s swaplog (%d entries)\n",
-		rb->sd->path, rb->n_read);
-	    fclose(rb->log);
-	    rb->log = NULL;
-	    store_dirs_rebuilding--;
-	    storeAufsDirCloseTmpSwapLog(rb->sd);
-	    storeRebuildComplete(&rb->counts);
-	    cbdataFree(rb);
+	    storeAufsDirRebuildComplete(rb);
 	    return;
 	}
 	rb->n_read++;
-	if (s.op <= SWAP_LOG_NOP)
-	    continue;
-	if (s.op >= SWAP_LOG_MAX)
-	    continue;
 	/*
 	 * BC: during 2.4 development, we changed the way swap file
 	 * numbers are assigned and stored.  The high 16 bits used
@@ -557,7 +597,8 @@ storeAufsDirRebuildFromSwapLog(void *data)
 	if (s.op == SWAP_LOG_ADD) {
 	    (void) 0;
 	} else if (s.op == SWAP_LOG_DEL) {
-	    if ((e = storeGet(s.key)) != NULL) {
+	    /* Delete unless we already have a newer copy */
+	    if ((e = storeGet(s.key)) != NULL && s.lastref >= e->lastref) {
 		/*
 		 * Make sure we don't unlink the file, it might be
 		 * in use by a subsequent entry.  Also note that
@@ -565,16 +606,7 @@ storeAufsDirRebuildFromSwapLog(void *data)
 		 * because adding to store_swap_size happens in
 		 * the cleanup procedure.
 		 */
-		storeExpireNow(e);
-		storeReleaseRequest(e);
-		if (e->swap_filen > -1) {
-		    storeAufsDirReplRemove(e);
-		    storeAufsDirMapBitReset(SD, e->swap_filen);
-		    e->swap_filen = -1;
-		    e->swap_dirn = -1;
-		}
-		storeRelease(e);
-		rb->counts.objcount--;
+		storeRecycle(e);
 		rb->counts.cancelcount++;
 	    }
 	    continue;
@@ -640,9 +672,7 @@ storeAufsDirRebuildFromSwapLog(void *data)
 	     * swapfiles back to StoreEntrys, we don't know the state
 	     * of the entry using that file.  */
 	    /* We'll assume the existing entry is valid, probably because
-	     * were in a slow rebuild and the the swap file number got taken
-	     * and the validation procedure hasn't run. */
-	    assert(rb->flags.need_to_validate);
+	     * the swap file number got taken while we rebuild */
 	    rb->counts.clashcount++;
 	    continue;
 	} else if (e && !disk_entry_newer) {
@@ -653,16 +683,7 @@ storeAufsDirRebuildFromSwapLog(void *data)
 	} else if (e) {
 	    /* key already exists, this swapfile not being used */
 	    /* junk old, load new */
-	    storeExpireNow(e);
-	    storeReleaseRequest(e);
-	    if (e->swap_filen > -1) {
-		storeAufsDirReplRemove(e);
-		/* Make sure we don't actually unlink the file */
-		storeAufsDirMapBitReset(SD, e->swap_filen);
-		e->swap_filen = -1;
-		e->swap_dirn = -1;
-	    }
-	    storeRelease(e);
+	    storeRecycle(e);
 	    rb->counts.dupcount++;
 	} else {
 	    /* URL doesnt exist, swapfile not in use */
@@ -686,6 +707,196 @@ storeAufsDirRebuildFromSwapLog(void *data)
     eventAdd("storeRebuild", storeAufsDirRebuildFromSwapLog, rb, 0.0, 1);
 }
 
+#if SIZEOF_SQUID_FILE_SZ != SIZEOF_SIZE_T
+/* This is an exact copy of the above, but using storeSwapLogDataOld entry type */
+static void
+storeAufsDirRebuildFromSwapLogOld(void *data)
+{
+    RebuildState *rb = data;
+    SwapDir *SD = rb->sd;
+    StoreEntry *e = NULL;
+    storeSwapLogDataOld s;
+    size_t ss = sizeof(storeSwapLogDataOld);
+    int count;
+    int used;			/* is swapfile already in use? */
+    int disk_entry_newer;	/* is the log entry newer than current entry? */
+    double x;
+    assert(rb != NULL);
+    /* load a number of objects per invocation */
+    for (count = 0; count < rb->speed; count++) {
+	if (fread(&s, ss, 1, rb->log) != 1) {
+	    storeAufsDirRebuildComplete(rb);
+	    return;
+	}
+	rb->n_read++;
+	/*
+	 * BC: during 2.4 development, we changed the way swap file
+	 * numbers are assigned and stored.  The high 16 bits used
+	 * to encode the SD index number.  There used to be a call
+	 * to storeDirProperFileno here that re-assigned the index 
+	 * bits.  Now, for backwards compatibility, we just need
+	 * to mask it off.
+	 */
+	s.swap_filen &= 0x00FFFFFF;
+	debug(47, 3) ("storeAufsDirRebuildFromSwapLog: %s %s %08X\n",
+	    swap_log_op_str[(int) s.op],
+	    storeKeyText(s.key),
+	    s.swap_filen);
+	if (s.op == SWAP_LOG_ADD) {
+	    (void) 0;
+	} else if (s.op == SWAP_LOG_DEL) {
+	    /* Delete unless we already have a newer copy */
+	    if ((e = storeGet(s.key)) != NULL && s.lastref >= e->lastref) {
+		/*
+		 * Make sure we don't unlink the file, it might be
+		 * in use by a subsequent entry.  Also note that
+		 * we don't have to subtract from store_swap_size
+		 * because adding to store_swap_size happens in
+		 * the cleanup procedure.
+		 */
+		storeRecycle(e);
+		rb->counts.cancelcount++;
+	    }
+	    continue;
+	} else {
+	    x = log(++rb->counts.bad_log_op) / log(10.0);
+	    if (0.0 == x - (double) (int) x)
+		debug(47, 1) ("WARNING: %d invalid swap log entries found\n",
+		    rb->counts.bad_log_op);
+	    rb->counts.invalid++;
+	    continue;
+	}
+	if ((++rb->counts.scancount & 0xFFF) == 0) {
+	    struct stat sb;
+	    if (0 == fstat(fileno(rb->log), &sb))
+		storeRebuildProgress(SD->index,
+		    (int) sb.st_size / ss, rb->n_read);
+	}
+	if (!storeAufsDirValidFileno(SD, s.swap_filen, 0)) {
+	    rb->counts.invalid++;
+	    continue;
+	}
+	if (EBIT_TEST(s.flags, KEY_PRIVATE)) {
+	    rb->counts.badflags++;
+	    continue;
+	}
+	e = storeGet(s.key);
+	used = storeAufsDirMapBitTest(SD, s.swap_filen);
+	/* If this URL already exists in the cache, does the swap log
+	 * appear to have a newer entry?  Compare 'lastref' from the
+	 * swap log to e->lastref. */
+	disk_entry_newer = e ? (s.lastref > e->lastref ? 1 : 0) : 0;
+	if (used && !disk_entry_newer) {
+	    /* log entry is old, ignore it */
+	    rb->counts.clashcount++;
+	    continue;
+	} else if (used && e && e->swap_filen == s.swap_filen && e->swap_dirn == SD->index) {
+	    /* swapfile taken, same URL, newer, update meta */
+	    if (e->store_status == STORE_OK) {
+		e->lastref = s.timestamp;
+		e->timestamp = s.timestamp;
+		e->expires = s.expires;
+		e->lastmod = s.lastmod;
+		e->flags = s.flags;
+		e->refcount += s.refcount;
+		storeAufsDirUnrefObj(SD, e);
+	    } else {
+		debug_trap("storeAufsDirRebuildFromSwapLog: bad condition");
+		debug(47, 1) ("\tSee %s:%d\n", __FILE__, __LINE__);
+	    }
+	    continue;
+	} else if (used) {
+	    /* swapfile in use, not by this URL, log entry is newer */
+	    /* This is sorta bad: the log entry should NOT be newer at this
+	     * point.  If the log is dirty, the filesize check should have
+	     * caught this.  If the log is clean, there should never be a
+	     * newer entry. */
+	    debug(47, 1) ("WARNING: newer swaplog entry for dirno %d, fileno %08X\n",
+		SD->index, s.swap_filen);
+	    /* I'm tempted to remove the swapfile here just to be safe,
+	     * but there is a bad race condition in the NOVM version if
+	     * the swapfile has recently been opened for writing, but
+	     * not yet opened for reading.  Because we can't map
+	     * swapfiles back to StoreEntrys, we don't know the state
+	     * of the entry using that file.  */
+	    /* We'll assume the existing entry is valid, probably because
+	     * the swap file number got taken while we rebuld */
+	    rb->counts.clashcount++;
+	    continue;
+	} else if (e && !disk_entry_newer) {
+	    /* key already exists, current entry is newer */
+	    /* keep old, ignore new */
+	    rb->counts.dupcount++;
+	    continue;
+	} else if (e) {
+	    /* key already exists, this swapfile not being used */
+	    /* junk old, load new */
+	    storeRecycle(e);
+	    rb->counts.dupcount++;
+	} else {
+	    /* URL doesnt exist, swapfile not in use */
+	    /* load new */
+	    (void) 0;
+	}
+	/* update store_swap_size */
+	rb->counts.objcount++;
+	e = storeAufsDirAddDiskRestore(SD, s.key,
+	    s.swap_filen,
+	    s.swap_file_sz,
+	    s.expires,
+	    s.timestamp,
+	    s.lastref,
+	    s.lastmod,
+	    s.refcount,
+	    s.flags,
+	    (int) rb->flags.clean);
+	storeDirSwapLog(e, SWAP_LOG_ADD);
+    }
+    eventAdd("storeRebuild", storeAufsDirRebuildFromSwapLogOld, rb, 0.0, 1);
+}
+
+#endif
+
+static void
+storeAufsDirRebuildFromSwapLogCheckVersion(void *data)
+{
+    RebuildState *rb = data;
+    storeSwapLogHeader hdr;
+
+    if (fread(&hdr, sizeof(hdr), 1, rb->log) != 1) {
+	storeAufsDirRebuildComplete(rb);
+	return;
+    }
+    if (hdr.op == SWAP_LOG_VERSION) {
+	if (fseek(rb->log, hdr.record_size, SEEK_SET) != 0) {
+	    storeAufsDirRebuildComplete(rb);
+	    return;
+	}
+	if (hdr.version == 1 && hdr.record_size == sizeof(storeSwapLogData)) {
+	    eventAdd("storeRebuild", storeAufsDirRebuildFromSwapLog, rb, 0.0, 1);
+	    return;
+	}
+#if SIZEOF_SQUID_FILE_SZ != SIZEOF_SIZE_T
+	if (hdr.version == 1 && hdr.record_size == sizeof(storeSwapLogDataOld)) {
+	    debug(47, 1) ("storeAufsDirRebuildFromSwapLog: Found current version but without large file support. Upgrading\n");
+	    eventAdd("storeRebuild", storeAufsDirRebuildFromSwapLogOld, rb, 0.0, 1);
+	    return;
+	}
+#endif
+	debug(47, 1) ("storeAufsDirRebuildFromSwapLog: Unsupported swap.state version %d size %d\n",
+	    hdr.version, hdr.record_size);
+	storeAufsDirRebuildComplete(rb);
+	return;
+    }
+    rewind(rb->log);
+    debug(47, 1) ("storeAufsDirRebuildFromSwapLog: Old version detected. Upgrading\n");
+#if SIZEOF_SQUID_FILE_SZ == SIZEOF_SIZE_T
+    eventAdd("storeRebuild", storeAufsDirRebuildFromSwapLog, rb, 0.0, 1);
+#else
+    eventAdd("storeRebuild", storeAufsDirRebuildFromSwapLogOld, rb, 0.0, 1);
+#endif
+}
+
 static int
 storeAufsDirGetNextFile(RebuildState * rb, sfileno * filn_p, int *size)
 {
@@ -778,7 +989,7 @@ storeAufsDirGetNextFile(RebuildState * rb, sfileno * filn_p, int *size)
 static StoreEntry *
 storeAufsDirAddDiskRestore(SwapDir * SD, const cache_key * key,
     sfileno file_number,
-    size_t swap_file_sz,
+    squid_file_sz swap_file_sz,
     time_t expires,
     time_t timestamp,
     time_t lastref,
@@ -791,7 +1002,7 @@ storeAufsDirAddDiskRestore(SwapDir * SD, const cache_key * key,
     debug(47, 5) ("storeAufsAddDiskRestore: %s, fileno=%08X\n", storeKeyText(key), file_number);
     /* if you call this you'd better be sure file_number is not 
      * already in use! */
-    e = new_StoreEntry(STORE_ENTRY_WITHOUT_MEMOBJ, NULL, NULL);
+    e = new_StoreEntry(STORE_ENTRY_WITHOUT_MEMOBJ, NULL);
     e->store_status = STORE_OK;
     storeSetMemStatus(e, NOT_IN_MEMORY);
     e->swap_status = SWAPOUT_DONE;
@@ -842,12 +1053,10 @@ storeAufsDirRebuild(SwapDir * sd)
 	    fclose(fp);
 	func = storeAufsDirRebuildFromDirectory;
     } else {
-	func = storeAufsDirRebuildFromSwapLog;
+	func = storeAufsDirRebuildFromSwapLogCheckVersion;
 	rb->log = fp;
 	rb->flags.clean = (unsigned int) clean;
     }
-    if (!clean)
-	rb->flags.need_to_validate = 1;
     debug(47, 1) ("Rebuilding storage in %s (%s)\n",
 	sd->path, clean ? "CLEAN" : "DIRTY");
     store_dirs_rebuilding++;
@@ -862,12 +1071,6 @@ storeAufsDirCloseTmpSwapLog(SwapDir * sd)
     char *new_path = xstrdup(storeAufsDirSwapLogFile(sd, ".new"));
     int fd;
     file_close(aioinfo->swaplog_fd);
-#if defined (_SQUID_OS2_) || defined (_SQUID_CYGWIN_)
-    if (unlink(swaplog_path) < 0) {
-	debug(50, 0) ("%s: %s\n", swaplog_path, xstrerror());
-	fatal("storeAufsDirCloseTmpSwapLog: unlink failed");
-    }
-#endif
     if (xrename(new_path, swaplog_path) < 0) {
 	fatal("storeAufsDirCloseTmpSwapLog: rename failed");
     }
@@ -882,6 +1085,31 @@ storeAufsDirCloseTmpSwapLog(SwapDir * sd)
     debug(47, 3) ("Cache Dir #%d log opened on FD %d\n", sd->index, fd);
 }
 
+static void
+storeSwapLogDataFree(void *s)
+{
+    memFree(s, MEM_SWAP_LOG_DATA);
+}
+
+static void
+storeAufsWriteSwapLogheader(int fd)
+{
+    storeSwapLogHeader *hdr = memAllocate(MEM_SWAP_LOG_DATA);
+    hdr->op = SWAP_LOG_VERSION;
+    hdr->version = 1;
+    hdr->record_size = sizeof(storeSwapLogData);
+    /* The header size is a full log record to keep some level of backward
+     * compatibility even if the actual header is smaller
+     */
+    file_write(fd,
+	-1,
+	hdr,
+	sizeof(storeSwapLogData),
+	NULL,
+	NULL,
+	(FREE *) storeSwapLogDataFree);
+}
+
 static FILE *
 storeAufsDirOpenTmpSwapLog(SwapDir * sd, int *clean_flag, int *zero_flag)
 {
@@ -911,15 +1139,13 @@ storeAufsDirOpenTmpSwapLog(SwapDir * sd, int *clean_flag, int *zero_flag)
 	fatal("storeDirOpenTmpSwapLog: Failed to open swap log.");
     }
     aioinfo->swaplog_fd = fd;
+    storeAufsWriteSwapLogheader(fd);
     /* open a read-only stream of the old log */
-    fp = fopen(swaplog_path, "r");
+    fp = fopen(swaplog_path, "rb");
     if (fp == NULL) {
 	debug(50, 0) ("%s: %s\n", swaplog_path, xstrerror());
 	fatal("Failed to open swap log for reading");
     }
-#if defined(_SQUID_CYGWIN_)
-    setmode(fileno(fp), O_BINARY);
-#endif
     memset(&clean_sb, '\0', sizeof(struct stat));
     if (stat(clean_path, &clean_sb) < 0)
 	*clean_flag = 0;
@@ -939,7 +1165,7 @@ struct _clean_state {
     char *new;
     char *cln;
     char *outbuf;
-    off_t outbuf_offset;
+    int outbuf_offset;
     int fd;
     RemovalPolicyWalker *walker;
 };
@@ -962,10 +1188,15 @@ storeAufsDirWriteCleanStart(SwapDir * sd)
     state->new = xstrdup(storeAufsDirSwapLogFile(sd, ".clean"));
     state->fd = file_open(state->new, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY);
     if (state->fd < 0) {
+	debug(50, 0) ("storeDirWriteCleanStart: %s: open: %s\n",
+	    state->new, xstrerror());
+	debug(50, 0) ("storeDirWriteCleanStart: Current swap logfile "
+	    "not replaced.\n");
 	xfree(state->new);
 	xfree(state);
 	return -1;
     }
+    storeAufsWriteSwapLogheader(state->fd);
     state->cur = xstrdup(storeAufsDirSwapLogFile(sd, NULL));
     state->cln = xstrdup(storeAufsDirSwapLogFile(sd, ".last-clean"));
     state->outbuf = xcalloc(CLEAN_BUF_SZ, 1);
@@ -1015,7 +1246,7 @@ storeAufsDirWriteCleanEntry(SwapDir * sd, const StoreEntry * e)
     s.swap_file_sz = e->swap_file_sz;
     s.refcount = e->refcount;
     s.flags = e->flags;
-    xmemcpy(&s.key, e->hash.key, MD5_DIGEST_CHARS);
+    xmemcpy(&s.key, e->hash.key, SQUID_MD5_DIGEST_LENGTH);
     xmemcpy(state->outbuf + state->outbuf_offset, &s, ss);
     state->outbuf_offset += ss;
     /* buffered write */
@@ -1065,12 +1296,9 @@ storeAufsDirWriteCleanDone(SwapDir * sd)
     fd = state->fd;
     /* rename */
     if (state->fd >= 0) {
-#if defined(_SQUID_OS2_) || defined (_SQUID_CYGWIN_)
+#if defined(_SQUID_OS2_) || defined(_SQUID_WIN32_)
 	file_close(state->fd);
 	state->fd = -1;
-	if (unlink(state->cur) < 0)
-	    debug(50, 0) ("storeDirWriteCleanLogs: unlinkd failed: %s, %s\n",
-		xstrerror(), state->cur);
 #endif
 	xrename(state->new, state->cur);
     }
@@ -1093,12 +1321,6 @@ storeAufsDirWriteCleanDone(SwapDir * sd)
     sd->log.clean.write = NULL;
 }
 
-static void
-storeSwapLogDataFree(void *s)
-{
-    memFree(s, MEM_SWAP_LOG_DATA);
-}
-
 static void
 storeAufsDirSwapLog(const SwapDir * sd, const StoreEntry * e, int op)
 {
@@ -1113,7 +1335,7 @@ storeAufsDirSwapLog(const SwapDir * sd, const StoreEntry * e, int op)
     s->swap_file_sz = e->swap_file_sz;
     s->refcount = e->refcount;
     s->flags = e->flags;
-    xmemcpy(s->key, e->hash.key, MD5_DIGEST_CHARS);
+    xmemcpy(s->key, e->hash.key, SQUID_MD5_DIGEST_LENGTH);
     file_write(aioinfo->swaplog_fd,
 	-1,
 	s,
@@ -1173,7 +1395,11 @@ storeAufsDirClean(int swap_index)
     if (dp == NULL) {
 	if (errno == ENOENT) {
 	    debug(36, 0) ("storeDirClean: WARNING: Creating %s\n", p1);
+#ifdef _SQUID_MSWIN_
+	    if (mkdir(p1) == 0)
+#else
 	    if (mkdir(p1, 0777) == 0)
+#endif
 		return 0;
 	}
 	debug(50, 0) ("storeDirClean: %s: %s\n", p1, xstrerror());
@@ -1218,55 +1444,38 @@ storeAufsDirClean(int swap_index)
 static void
 storeAufsDirCleanEvent(void *unused)
 {
-    static int swap_index = 0;
-    int i;
+    static int swap_index = -1;
     int j = 0;
     int n = 0;
     /*
      * Assert that there are AUFS cache_dirs configured, otherwise
      * we should never be called.
      */
-    assert(n_asyncufs_dirs);
-    if (NULL == asyncufs_dir_index) {
+    if (swap_index == -1) {
 	SwapDir *sd;
 	squidaioinfo_t *aioinfo;
-	/*
-	 * Initialize the little array that translates AUFS cache_dir
-	 * number into the Config.cacheSwap.swapDirs array index.
-	 */
-	asyncufs_dir_index = xcalloc(n_asyncufs_dirs, sizeof(*asyncufs_dir_index));
-	for (i = 0, n = 0; i < Config.cacheSwap.n_configured; i++) {
-	    sd = &Config.cacheSwap.swapDirs[i];
-	    if (!storeAufsDirIs(sd))
-		continue;
-	    asyncufs_dir_index[n++] = i;
-	    aioinfo = (squidaioinfo_t *) sd->fsdata;
-	    j += (aioinfo->l1 * aioinfo->l2);
-	}
-	assert(n == n_asyncufs_dirs);
 	/*
 	 * Start the storeAufsDirClean() swap_index with a random
 	 * value.  j equals the total number of AUFS level 2
 	 * swap directories
 	 */
+	for (n = 0; n < n_asyncufs_dirs; n++) {
+	    sd = &Config.cacheSwap.swapDirs[asyncufs_dir_index[n]];
+	    aioinfo = (squidaioinfo_t *) sd->fsdata;
+	    j += (aioinfo->l1 * aioinfo->l2);
+	}
 	swap_index = (int) (squid_random() % j);
     }
     if (0 == store_dirs_rebuilding) {
 	n = storeAufsDirClean(swap_index);
 	swap_index++;
+	if (swap_index < 0)
+	    swap_index = 0;
     }
     eventAdd("storeDirClean", storeAufsDirCleanEvent, NULL,
 	15.0 * exp(-0.25 * n), 1);
 }
 
-static int
-storeAufsDirIs(SwapDir * sd)
-{
-    if (strncmp(sd->type, "aufs", 4) == 0)
-	return 1;
-    return 0;
-}
-
 /*
  * Does swapfile number 'fn' belong in cachedir #F0,
  * level1 dir #F1, level2 dir #F2?
@@ -1310,6 +1519,7 @@ storeAufsDirValidFileno(SwapDir * SD, sfileno filn, int flag)
 void
 storeAufsDirMaintain(SwapDir * SD)
 {
+    squidaioinfo_t *aioinfo = (squidaioinfo_t *) SD->fsdata;
     StoreEntry *e = NULL;
     int removed = 0;
     int max_scan;
@@ -1332,7 +1542,7 @@ storeAufsDirMaintain(SwapDir * SD)
 	f, max_scan, max_remove);
     walker = SD->repl->PurgeInit(SD->repl, max_scan);
     while (1) {
-	if (SD->cur_size < SD->low_size)
+	if (SD->cur_size < SD->low_size && aioinfo->map->n_files_in_map < FILEMAP_MAX)
 	    break;
 	if (removed >= max_remove)
 	    break;
@@ -1341,6 +1551,8 @@ storeAufsDirMaintain(SwapDir * SD)
 	    break;		/* no more objects */
 	removed++;
 	storeRelease(e);
+	if (aioQueueSize() > MAGIC2)
+	    break;
     }
     walker->Done(walker);
     debug(47, (removed ? 2 : 3)) ("storeAufsDirMaintain: %s removed %d/%d f=%.03f max_scan=%d\n",
@@ -1357,22 +1569,19 @@ storeAufsDirMaintain(SwapDir * SD)
 int
 storeAufsDirCheckObj(SwapDir * SD, const StoreEntry * e)
 {
-    int loadav;
-    int ql;
+    return 1;
+}
+
+int
+storeAufsDirCheckLoadAv(SwapDir * SD, store_op_t op)
+{
+    int loadav, ql;
 
-#if OLD_UNUSED_CODE
-    if (storeAufsDirExpiredReferenceAge(SD) < 300) {
-	debug(47, 3) ("storeAufsDirCheckObj: NO: LRU Age = %d\n",
-	    storeAufsDirExpiredReferenceAge(SD));
-	/* store_check_cachable_hist.no.lru_age_too_low++; */
-	return -1;
-    }
-#endif
     ql = aioQueueSize();
-    if (ql == 0)
-	loadav = 0;
-    loadav = ql * 1000 / MAGIC1;
-    debug(47, 9) ("storeAufsDirCheckObj: load=%d\n", loadav);
+    if (ql == 0) {
+	return AUFS_LOAD_BASE;
+    }
+    loadav = AUFS_LOAD_BASE + (ql * AUFS_LOAD_QUEUE_WEIGHT / MAGIC1);
     return loadav;
 }
 
@@ -1417,8 +1626,8 @@ storeAufsDirUnlinkFile(SwapDir * SD, sfileno f)
 {
     debug(79, 3) ("storeAufsDirUnlinkFile: unlinking fileno %08X\n", f);
     /* storeAufsDirMapBitReset(SD, f); */
-#if USE_TRUNCATE_NOT_UNLINK
-    aioTruncate(storeAufsDirFullPath(SD, f, NULL), NULL, NULL);
+#if USE_TRUNCATE
+    aioTruncate(storeAufsDirFullPath(SD, f, NULL), 0, NULL, NULL);
 #else
     aioUnlink(storeAufsDirFullPath(SD, f, NULL), NULL, NULL);
 #endif
@@ -1455,10 +1664,17 @@ void
 storeAufsDirStats(SwapDir * SD, StoreEntry * sentry)
 {
     squidaioinfo_t *aioinfo = SD->fsdata;
+#ifdef HAVE_STATVFS
+    fsblkcnt_t totl_kb;
+    fsblkcnt_t free_kb;
+    fsfilcnt_t totl_in;
+    fsfilcnt_t free_in;
+#else
     int totl_kb = 0;
     int free_kb = 0;
     int totl_in = 0;
     int free_in = 0;
+#endif
     int x;
     storeAppendPrintf(sentry, "First level subdirectories: %d\n", aioinfo->l1);
     storeAppendPrintf(sentry, "Second level subdirectories: %d\n", aioinfo->l2);
@@ -1466,11 +1682,22 @@ storeAufsDirStats(SwapDir * SD, StoreEntry * sentry)
     storeAppendPrintf(sentry, "Current Size: %d KB\n", SD->cur_size);
     storeAppendPrintf(sentry, "Percent Used: %0.2f%%\n",
 	100.0 * SD->cur_size / SD->max_size);
+    storeAppendPrintf(sentry, "Current load metric: %d / %d\n", storeAufsDirCheckLoadAv(SD, ST_OP_CREATE), MAX_LOAD_VALUE);
     storeAppendPrintf(sentry, "Filemap bits in use: %d of %d (%d%%)\n",
 	aioinfo->map->n_files_in_map, aioinfo->map->max_n_files,
 	percent(aioinfo->map->n_files_in_map, aioinfo->map->max_n_files));
     x = storeDirGetUFSStats(SD->path, &totl_kb, &free_kb, &totl_in, &free_in);
     if (0 == x) {
+#ifdef HAVE_STATVFS
+	storeAppendPrintf(sentry, "Filesystem Space in use: %" PRIu64 "/%" PRIu64 " KB (%.0f%%)\n",
+	    (uint64_t) (totl_kb - free_kb),
+	    (uint64_t) totl_kb,
+	    dpercent(totl_kb - free_kb, totl_kb));
+	storeAppendPrintf(sentry, "Filesystem Inodes in use: %" PRIu64 "/%" PRIu64 " (%.0f%%)\n",
+	    (uint64_t) (totl_in - free_in),
+	    (uint64_t) totl_in,
+	    dpercent(totl_in - free_in, totl_in));
+#else
 	storeAppendPrintf(sentry, "Filesystem Space in use: %d/%d KB (%d%%)\n",
 	    totl_kb - free_kb,
 	    totl_kb,
@@ -1479,6 +1706,7 @@ storeAufsDirStats(SwapDir * SD, StoreEntry * sentry)
 	    totl_in - free_in,
 	    totl_in,
 	    percent(totl_in - free_in, totl_in));
+#endif
     }
     storeAppendPrintf(sentry, "Flags:");
     if (SD->flags.selected)
@@ -1651,6 +1879,7 @@ storeAufsDirParse(SwapDir * sd, int index, char *path)
     aioinfo->swaplog_fd = -1;
     aioinfo->map = NULL;	/* Debugging purposes */
     aioinfo->suggest = 0;
+    sd->checkconfig = storeAufsCheckConfig;
     sd->init = storeAufsDirInit;
     sd->newfs = storeAufsDirNewfs;
     sd->dump = storeAufsDirDump;
@@ -1659,6 +1888,7 @@ storeAufsDirParse(SwapDir * sd, int index, char *path)
     sd->statfs = storeAufsDirStats;
     sd->maintainfs = storeAufsDirMaintain;
     sd->checkobj = storeAufsDirCheckObj;
+    sd->checkload = storeAufsDirCheckLoadAv;
     sd->refobj = storeAufsDirRefObj;
     sd->unrefobj = storeAufsDirUnrefObj;
     sd->callback = aioCheckCallbacks;
@@ -1669,6 +1899,7 @@ storeAufsDirParse(SwapDir * sd, int index, char *path)
     sd->obj.read = storeAufsRead;
     sd->obj.write = storeAufsWrite;
     sd->obj.unlink = storeAufsUnlink;
+    sd->obj.recycle = storeAufsRecycle;
     sd->log.open = storeAufsDirOpenSwapLog;
     sd->log.close = storeAufsDirCloseSwapLog;
     sd->log.write = storeAufsDirSwapLog;
@@ -1680,6 +1911,9 @@ storeAufsDirParse(SwapDir * sd, int index, char *path)
 
     /* Initialise replacement policy stuff */
     sd->repl = createRemovalPolicy(Config.replPolicy);
+
+    asyncufs_dir_index = realloc(asyncufs_dir_index, (n_asyncufs_dirs + 1) * sizeof(*asyncufs_dir_index));
+    asyncufs_dir_index[n_asyncufs_dirs++] = index;
 }
 
 /*
