@@ -1,6 +1,6 @@
 
 /*
- * $Id: net_db.c,v 1.158.2.1 2002/06/26 09:57:25 hno Exp $
+ * $Id: net_db.c,v 1.176.2.2 2008/04/08 22:52:59 hno Exp $
  *
  * DEBUG: section 38    Network Measurement Database
  * AUTHOR: Duane Wessels
@@ -42,8 +42,8 @@ typedef struct {
     StoreEntry *e;
     store_client *sc;
     request_t *r;
-    off_t seen;
-    off_t used;
+    squid_off_t seen;
+    squid_off_t used;
     size_t buf_sz;
     char *buf;
 } netdbExchangeState;
@@ -70,6 +70,9 @@ static FREE netdbFreeNameEntry;
 static FREE netdbFreeNetdbEntry;
 static STCB netdbExchangeHandleReply;
 static void netdbExchangeDone(void *);
+#if USE_ICMP
+void netdbDump(StoreEntry * sentry);
+#endif
 
 /* We have to keep a local list of peer names.  The Peers structure
  * gets freed during a reconfigure.  We want this database to
@@ -247,7 +250,7 @@ netdbSendPing(const ipcache_addrs * ia, void *data)
 	    hostname, n->network, na->network);
 	x = (net_db_name *) hash_lookup(host_table, hostname);
 	if (x == NULL) {
-	    debug(38, 1) ("netdbSendPing: net_db_name list bug: %s not found", hostname);
+	    debug(38, 1) ("netdbSendPing: net_db_name list bug: %s not found\n", hostname);
 	    xfree(hostname);
 	    return;
 	}
@@ -367,13 +370,13 @@ sortPeerByRtt(const void *A, const void *B)
 static void
 netdbSaveState(void *foo)
 {
-    LOCAL_ARRAY(char, path, SQUID_MAXPATHLEN);
     Logfile *lf;
     netdbEntry *n;
     net_db_name *x;
     struct timeval start = current_time;
     int count = 0;
-    snprintf(path, SQUID_MAXPATHLEN, "%s/netdb_state", storeSwapDir(0));
+    if (strcmp(Config.netdbFilename, "none") == 0)
+	return;
     /*
      * This was nicer when we were using stdio, but thanks to
      * Solaris bugs, its a bad idea.  fopen can fail if more than
@@ -383,10 +386,10 @@ netdbSaveState(void *foo)
      * unlink() is here because there is currently no way to make
      * logfileOpen() use O_TRUNC.
      */
-    unlink(path);
-    lf = logfileOpen(path, 4096, 0);
+    unlink(Config.netdbFilename);
+    lf = logfileOpen(Config.netdbFilename, 4096, 0);
     if (NULL == lf) {
-	debug(50, 1) ("netdbSaveState: %s: %s\n", path, xstrerror());
+	debug(50, 1) ("netdbSaveState: %s: %s\n", Config.netdbFilename, xstrerror());
 	return;
     }
     hash_first(addr_table);
@@ -417,7 +420,6 @@ netdbSaveState(void *foo)
 static void
 netdbReloadState(void)
 {
-    LOCAL_ARRAY(char, path, SQUID_MAXPATHLEN);
     char *buf;
     char *t;
     char *s;
@@ -429,21 +431,24 @@ netdbReloadState(void)
     struct in_addr addr;
     int count = 0;
     struct timeval start = current_time;
-    snprintf(path, SQUID_MAXPATHLEN, "%s/netdb_state", storeSwapDir(0));
+
+    if (strcmp(Config.netdbFilename, "none") == 0)
+	return;
+
     /*
      * This was nicer when we were using stdio, but thanks to
      * Solaris bugs, its a bad idea.  fopen can fail if more than
      * 256 FDs are open.
      */
-    fd = file_open(path, O_RDONLY | O_TEXT);
+    fd = file_open(Config.netdbFilename, O_RDONLY | O_BINARY);
     if (fd < 0)
 	return;
     if (fstat(fd, &sb) < 0) {
 	file_close(fd);
 	return;
     }
-    t = buf = xcalloc(1, sb.st_size + 1);
-    l = FD_READ_METHOD(fd, buf, sb.st_size);
+    t = buf = xcalloc(1, (size_t) sb.st_size + 1);
+    l = FD_READ_METHOD(fd, buf, (int) sb.st_size);
     file_close(fd);
     if (l <= 0)
 	return;
@@ -531,7 +536,7 @@ netdbExchangeHandleReply(void *data, char *buf, ssize_t size)
 {
     netdbExchangeState *ex = data;
     int rec_sz = 0;
-    off_t o;
+    ssize_t o;
     struct in_addr addr;
     double rtt;
     double hops;
@@ -555,30 +560,28 @@ netdbExchangeHandleReply(void *data, char *buf, ssize_t size)
     p = buf;
     if (0 == ex->used) {
 	/* skip reply headers */
-	if ((hdr_sz = headersEnd(p, size))) {
-	    debug(38, 5) ("netdbExchangeHandleReply: hdr_sz = %d\n", hdr_sz);
-	    rep = ex->e->mem_obj->reply;
-	    if (0 == rep->sline.status)
-		httpReplyParse(rep, buf, hdr_sz);
-	    debug(38, 3) ("netdbExchangeHandleReply: reply status %d\n",
-		rep->sline.status);
-	    if (HTTP_OK != rep->sline.status) {
-		netdbExchangeDone(ex);
-		return;
-	    }
-	    assert(size >= hdr_sz);
-	    ex->used += hdr_sz;
+	rep = ex->e->mem_obj->reply;
+	hdr_sz = rep->hdr_sz;
+	debug(38, 5) ("netdbExchangeHandleReply: hdr_sz = %ld\n", (long int) hdr_sz);
+	debug(38, 3) ("netdbExchangeHandleReply: reply status %d\n",
+	    rep->sline.status);
+	if (HTTP_OK != rep->sline.status) {
+	    netdbExchangeDone(ex);
+	    return;
+	}
+	ex->used += hdr_sz;
+	if (size < hdr_sz) {
 	    size -= hdr_sz;
 	    p += hdr_sz;
 	} else {
 	    size = 0;
 	}
     }
-    debug(38, 5) ("netdbExchangeHandleReply: start parsing loop, size = %d\n",
-	size);
+    debug(38, 5) ("netdbExchangeHandleReply: start parsing loop, size = %ld\n",
+	(long int) size);
     while (size >= rec_sz) {
-	debug(38, 5) ("netdbExchangeHandleReply: in parsing loop, size = %d\n",
-	    size);
+	debug(38, 5) ("netdbExchangeHandleReply: in parsing loop, size = %ld\n",
+	    (long int) size);
 	addr.s_addr = any_addr.s_addr;
 	hops = rtt = 0.0;
 	for (o = 0; o < rec_sz;) {
@@ -621,7 +624,7 @@ netdbExchangeHandleReply(void *data, char *buf, ssize_t size)
     }
     debug(38, 3) ("netdbExchangeHandleReply: used %d entries, (x %d bytes) == %d bytes total\n",
 	nused, rec_sz, nused * rec_sz);
-    debug(38, 3) ("netdbExchangeHandleReply: seen %d, used %d\n", ex->seen, ex->used);
+    debug(38, 3) ("netdbExchangeHandleReply: seen %ld, used %ld\n", (long int) ex->seen, (long int) ex->used);
     if (EBIT_TEST(ex->e->flags, ENTRY_ABORTED)) {
 	debug(38, 3) ("netdbExchangeHandleReply: ENTRY_ABORTED\n");
 	netdbExchangeDone(ex);
@@ -646,7 +649,7 @@ netdbExchangeDone(void *data)
     debug(38, 3) ("netdbExchangeDone: %s\n", storeUrl(ex->e));
     memFree(ex->buf, MEM_4K_BUF);
     requestUnlink(ex->r);
-    storeUnregister(ex->sc, ex->e, ex);
+    storeClientUnregister(ex->sc, ex->e, ex);
     storeUnlockObject(ex->e);
     cbdataUnlock(ex->p);
     cbdataFree(ex);
@@ -741,10 +744,10 @@ netdbHops(struct in_addr addr)
     return 256;
 }
 
+#if USE_ICMP
 void
 netdbDump(StoreEntry * sentry)
 {
-#if USE_ICMP
     netdbEntry *n;
     netdbEntry **list;
     net_db_name *x;
@@ -791,11 +794,8 @@ netdbDump(StoreEntry * sentry)
 	}
     }
     xfree(list);
-#else
-    storeAppendPrintf(sentry,
-	"NETDB support not compiled into this Squid cache.\n");
-#endif
 }
+#endif
 
 int
 netdbHostHops(const char *host)
@@ -833,23 +833,10 @@ netdbHostData(const char *host, int *samp, int *rtt, int *hops)
     *samp = n->pings_recv;
     *rtt = (int) (n->rtt + 0.5);
     *hops = (int) (n->hops + 0.5);
+    n->last_use_time = squid_curtime;
 #endif
 }
 
-int
-netdbHostPeerRtt(const char *host, peer * p)
-{
-#if USE_ICMP
-    const netdbEntry *n = netdbLookupHost(host);
-    if (n) {
-	const net_db_peer *np = netdbPeerByName(n, p->host);
-	if (np && np->expires >= squid_curtime)
-	    return (int) (np->rtt + 0.5);
-    }
-#endif
-    return 0;
-}
-
 void
 netdbUpdatePeer(request_t * r, peer * e, int irtt, int ihops)
 {
@@ -920,7 +907,6 @@ void
 netdbBinaryExchange(StoreEntry * s)
 {
     http_reply *reply = s->mem_obj->reply;
-    http_version_t version;
 #if USE_ICMP
     netdbEntry *n;
     int i;
@@ -930,9 +916,7 @@ netdbBinaryExchange(StoreEntry * s)
     struct in_addr addr;
     storeBuffer(s);
     httpReplyReset(reply);
-    httpBuildVersion(&version, 1, 0);
-    httpReplySetHeaders(reply, version, HTTP_OK, "OK",
-	NULL, -1, squid_curtime, -2);
+    httpReplySetHeaders(reply, HTTP_OK, "OK", NULL, -1, squid_curtime, -1);
     httpReplySwapOut(reply, s);
     rec_sz = 0;
     rec_sz += 1 + sizeof(addr.s_addr);
@@ -973,9 +957,8 @@ netdbBinaryExchange(StoreEntry * s)
     memFree(buf, MEM_4K_BUF);
 #else
     httpReplyReset(reply);
-    httpBuildVersion(&version, 1, 0);
-    httpReplySetHeaders(reply, version, HTTP_BAD_REQUEST, "Bad Request",
-	NULL, -1, squid_curtime, -2);
+    httpReplySetHeaders(reply, HTTP_BAD_REQUEST, "Bad Request", NULL, -1, -1, -1);
+    httpReplySwapOut(reply, s);
     storeAppendPrintf(s, "NETDB support not compiled into this Squid cache.\n");
 #endif
     storeComplete(s);
@@ -1007,11 +990,11 @@ netdbExchangeStart(void *data)
     requestLink(ex->r);
     assert(NULL != ex->r);
     httpBuildVersion(&ex->r->http_ver, 1, 0);
-    ex->e = storeCreateEntry(uri, uri, null_request_flags, METHOD_GET);
+    ex->e = storeCreateEntry(uri, null_request_flags, METHOD_GET);
     ex->buf_sz = 4096;
     ex->buf = memAllocate(MEM_4K_BUF);
     assert(NULL != ex->e);
-    ex->sc = storeClientListAdd(ex->e, ex);
+    ex->sc = storeClientRegister(ex->e, ex);
     storeClientCopy(ex->sc, ex->e, ex->seen, ex->used, ex->buf_sz,
 	ex->buf, netdbExchangeHandleReply, ex);
     ex->r->flags.loopdetect = 1;	/* cheat! -- force direct */
@@ -1041,6 +1024,7 @@ netdbClosestParent(request_t * request)
 	return NULL;
     if (0 == n->n_peers)
 	return NULL;
+    n->last_use_time = squid_curtime;
     /* 
      * Find the parent with the least RTT to the origin server.
      * Make sure we don't return a parent who is farther away than
