@@ -1,6 +1,6 @@
 
 /*
- * $Id: stat.c,v 1.351.2.1 2002/07/21 23:54:29 hno Exp $
+ * $Id: stat.c,v 1.384.2.1 2008/05/04 23:23:13 hno Exp $
  *
  * DEBUG: section 18    Cache Manager Statistics
  * AUTHOR: Harvest Derived
@@ -36,8 +36,6 @@
 
 #include "squid.h"
 
-#define DEBUG_OPENFD 1
-
 typedef int STOBJFLT(const StoreEntry *);
 typedef struct {
     StoreEntry *sentry;
@@ -58,14 +56,12 @@ static void statCountersInitSpecial(StatCounters *);
 static void statCountersClean(StatCounters *);
 static void statCountersCopy(StatCounters * dest, const StatCounters * orig);
 static double statMedianSvc(int, int);
-static void statStoreEntry(StoreEntry * s, StoreEntry * e);
+static void statStoreEntry(MemBuf * mb, StoreEntry * e);
 static double statCPUUsage(int minutes);
 static OBJH stat_io_get;
 static OBJH stat_objects_get;
 static OBJH stat_vmobjects_get;
-#if DEBUG_OPENFD
 static OBJH statOpenfdObj;
-#endif
 static EVH statObjects;
 static OBJH info_get;
 static OBJH statFiledescriptors;
@@ -179,16 +175,6 @@ stat_io_get(StoreEntry * sentry)
     }
 
     storeAppendPrintf(sentry, "\n");
-    storeAppendPrintf(sentry, "WAIS I/O\n");
-    storeAppendPrintf(sentry, "number of reads: %d\n", IOStats.Wais.reads);
-    storeAppendPrintf(sentry, "Read Histogram:\n");
-    for (i = 0; i < 16; i++) {
-	storeAppendPrintf(sentry, "%5d-%5d: %9d %2d%%\n",
-	    i ? (1 << (i - 1)) + 1 : 1,
-	    1 << i,
-	    IOStats.Wais.read_hist[i],
-	    percent(IOStats.Wais.read_hist[i], IOStats.Wais.reads));
-    }
 }
 
 static const char *
@@ -218,8 +204,8 @@ storeEntryFlags(const StoreEntry * entry)
 	strcat(buf, "DELAY_SENDING,");
     if (EBIT_TEST(flags, RELEASE_REQUEST))
 	strcat(buf, "RELEASE_REQUEST,");
-    if (EBIT_TEST(flags, REFRESH_REQUEST))
-	strcat(buf, "REFRESH_REQUEST,");
+    if (EBIT_TEST(flags, REFRESH_FAILURE))
+	strcat(buf, "REFRESH_FAILURE,");
     if (EBIT_TEST(flags, ENTRY_CACHABLE))
 	strcat(buf, "CACHABLE,");
     if (EBIT_TEST(flags, ENTRY_DISPATCHED))
@@ -254,55 +240,56 @@ describeTimestamps(const StoreEntry * entry)
 }
 
 static void
-statStoreEntry(StoreEntry * s, StoreEntry * e)
+statStoreEntry(MemBuf * mb, StoreEntry * e)
 {
     MemObject *mem = e->mem_obj;
     int i;
     struct _store_client *sc;
     dlink_node *node;
-    storeAppendPrintf(s, "KEY %s\n", storeKeyText(e->hash.key));
+    memBufPrintf(mb, "KEY %s\n", storeKeyText(e->hash.key));
+    /* XXX should this url be escaped? */
     if (mem)
-	storeAppendPrintf(s, "\t%s %s\n",
-	    RequestMethodStr[mem->method], mem->log_url);
-    storeAppendPrintf(s, "\t%s\n", describeStatuses(e));
-    storeAppendPrintf(s, "\t%s\n", storeEntryFlags(e));
-    storeAppendPrintf(s, "\t%s\n", describeTimestamps(e));
-    storeAppendPrintf(s, "\t%d locks, %d clients, %d refs\n",
+	memBufPrintf(mb, "\t%s %s\n",
+	    RequestMethods[mem->method].str, mem->url);
+    if (mem && mem->store_url)
+	memBufPrintf(mb, "\tStore lookup URL: %s\n", mem->store_url);
+    memBufPrintf(mb, "\t%s\n", describeStatuses(e));
+    memBufPrintf(mb, "\t%s\n", storeEntryFlags(e));
+    memBufPrintf(mb, "\t%s\n", describeTimestamps(e));
+    memBufPrintf(mb, "\t%d locks, %d clients, %d refs\n",
 	(int) e->lock_count,
 	storePendingNClients(e),
 	(int) e->refcount);
-    storeAppendPrintf(s, "\tSwap Dir %d, File %#08X\n",
+    memBufPrintf(mb, "\tSwap Dir %d, File %#08X\n",
 	e->swap_dirn, e->swap_filen);
     if (mem != NULL) {
-	storeAppendPrintf(s, "\tinmem_lo: %d\n", (int) mem->inmem_lo);
-	storeAppendPrintf(s, "\tinmem_hi: %d\n", (int) mem->inmem_hi);
-	storeAppendPrintf(s, "\tswapout: %d bytes queued\n",
-	    (int) mem->swapout.queue_offset);
+	memBufPrintf(mb, "\tinmem_lo: %" PRINTF_OFF_T "\n", mem->inmem_lo);
+	memBufPrintf(mb, "\tinmem_hi: %" PRINTF_OFF_T "\n", mem->inmem_hi);
+	memBufPrintf(mb, "\tswapout: %" PRINTF_OFF_T " bytes queued\n",
+	    mem->swapout.queue_offset);
 	if (mem->swapout.sio)
-	    storeAppendPrintf(s, "\tswapout: %d bytes written\n",
-		(int) storeOffset(mem->swapout.sio));
+	    memBufPrintf(mb, "\tswapout: %" PRINTF_OFF_T " bytes written\n",
+		storeOffset(mem->swapout.sio));
 	for (i = 0, node = mem->clients.head; node; node = node->next, i++) {
 	    sc = (store_client *) node->data;
-	    if (sc->callback_data == NULL)
-		continue;
-	    storeAppendPrintf(s, "\tClient #%d, %p\n", i, sc->callback_data);
-	    storeAppendPrintf(s, "\t\tcopy_offset: %d\n",
-		(int) sc->copy_offset);
-	    storeAppendPrintf(s, "\t\tseen_offset: %d\n",
-		(int) sc->seen_offset);
-	    storeAppendPrintf(s, "\t\tcopy_size: %d\n",
+	    memBufPrintf(mb, "\tClient #%d, %p\n", i, sc->callback_data);
+	    memBufPrintf(mb, "\t\tcopy_offset: %" PRINTF_OFF_T "\n",
+		sc->copy_offset);
+	    memBufPrintf(mb, "\t\tseen_offset: %" PRINTF_OFF_T "\n",
+		sc->seen_offset);
+	    memBufPrintf(mb, "\t\tcopy_size: %d\n",
 		(int) sc->copy_size);
-	    storeAppendPrintf(s, "\t\tflags:");
+	    memBufPrintf(mb, "\t\tflags:");
 	    if (sc->flags.disk_io_pending)
-		storeAppendPrintf(s, " disk_io_pending");
+		memBufPrintf(mb, " disk_io_pending");
 	    if (sc->flags.store_copying)
-		storeAppendPrintf(s, " store_copying");
+		memBufPrintf(mb, " store_copying");
 	    if (sc->flags.copy_event_pending)
-		storeAppendPrintf(s, " copy_event_pending");
-	    storeAppendPrintf(s, "\n");
+		memBufPrintf(mb, " copy_event_pending");
+	    memBufPrintf(mb, "\n");
 	}
     }
-    storeAppendPrintf(s, "\n");
+    memBufPrintf(mb, "\n");
 }
 
 /* process objects list */
@@ -326,19 +313,23 @@ statObjects(void *data)
 	eventAdd("statObjects", statObjects, state, 0.1, 1);
 	return;
     }
-    storeBuffer(state->sentry);
     debug(49, 3) ("statObjects: Bucket #%d\n", state->bucket);
     link_next = hash_get_bucket(store_table, state->bucket);
-    while (NULL != (link_ptr = link_next)) {
-	link_next = link_ptr->next;
-	e = (StoreEntry *) link_ptr;
-	if (state->filter && 0 == state->filter(e))
-	    continue;
-	statStoreEntry(state->sentry, e);
+    if (link_next) {
+	MemBuf mb;
+	memBufDefInit(&mb);
+	while (NULL != (link_ptr = link_next)) {
+	    link_next = link_ptr->next;
+	    e = (StoreEntry *) link_ptr;
+	    if (state->filter && 0 == state->filter(e))
+		continue;
+	    statStoreEntry(&mb, e);
+	}
+	storeAppend(state->sentry, mb.buf, mb.size);
+	memBufClean(&mb);
     }
     state->bucket++;
     eventAdd("statObjects", statObjects, state, 0.0, 1);
-    storeBufferFlush(state->sentry);
 }
 
 static void
@@ -370,7 +361,6 @@ stat_vmobjects_get(StoreEntry * sentry)
     statObjectsStart(sentry, statObjectsVmFilter);
 }
 
-#if DEBUG_OPENFD
 static int
 statObjectsOpenfdFilter(const StoreEntry * e)
 {
@@ -387,7 +377,35 @@ statOpenfdObj(StoreEntry * sentry)
     statObjectsStart(sentry, statObjectsOpenfdFilter);
 }
 
-#endif
+static int
+statObjectsPendingFilter(const StoreEntry * e)
+{
+    if (e->store_status != STORE_PENDING)
+	return 0;
+    return 1;
+}
+
+static void
+statPendingObj(StoreEntry * sentry)
+{
+    statObjectsStart(sentry, statObjectsPendingFilter);
+}
+
+static int
+statObjectsClientsFilter(const StoreEntry * e)
+{
+    if (e->mem_obj == NULL)
+	return 0;
+    if (e->mem_obj->clients.head == NULL)
+	return 0;
+    return 1;
+}
+
+static void
+statClientsObj(StoreEntry * sentry)
+{
+    statObjectsStart(sentry, statObjectsClientsFilter);
+}
 
 #ifdef XMALLOC_STATISTICS
 static void
@@ -405,7 +423,15 @@ fdRemoteAddr(const fde * f)
     LOCAL_ARRAY(char, buf, 32);
     if (f->type != FD_SOCKET)
 	return null_string;
-    snprintf(buf, 32, "%s.%d", f->ipaddr, (int) f->remote_port);
+    if (*f->ipaddr)
+	snprintf(buf, 32, "%s.%d", f->ipaddr, (int) f->remote_port);
+    else {
+	if (f->local_addr.s_addr != any_addr.s_addr) {
+	    snprintf(buf, 32, "%s.%d", inet_ntoa(f->local_addr), (int) f->local_port);
+	} else {
+	    snprintf(buf, 32, "*.%d", (int) f->local_port);
+	}
+    }
     return buf;
 }
 
@@ -415,23 +441,39 @@ statFiledescriptors(StoreEntry * sentry)
     int i;
     fde *f;
     storeAppendPrintf(sentry, "Active file descriptors:\n");
+#ifdef _SQUID_MSWIN_
+    storeAppendPrintf(sentry, "%-4s %-10s %-6s %-4s %-7s* %-7s* %-21s %s\n",
+	"File",
+	"Handle",
+#else
     storeAppendPrintf(sentry, "%-4s %-6s %-4s %-7s* %-7s* %-21s %s\n",
 	"File",
+#endif
 	"Type",
 	"Tout",
 	"Nread",
 	"Nwrite",
 	"Remote Address",
 	"Description");
+#ifdef _SQUID_MSWIN_
+    storeAppendPrintf(sentry, "---- ---------- ------ ---- -------- -------- --------------------- ------------------------------\n");
+#else
     storeAppendPrintf(sentry, "---- ------ ---- -------- -------- --------------------- ------------------------------\n");
+#endif
     for (i = 0; i < Squid_MaxFD; i++) {
 	f = &fd_table[i];
 	if (!f->flags.open)
 	    continue;
-	storeAppendPrintf(sentry, "%4d %-6.6s %4d %7d%c %7d%c %-21s %s\n",
+#ifdef _SQUID_MSWIN_
+	storeAppendPrintf(sentry, "%4d 0x%-8lX %-6.6s %4d %7" PRINTF_OFF_T "%c %7" PRINTF_OFF_T "%c %-21s %s\n",
 	    i,
+	    f->win32.handle,
+#else
+	storeAppendPrintf(sentry, "%4d %-6.6s %4d %7" PRINTF_OFF_T "%c %7" PRINTF_OFF_T "%c %-21s %s\n",
+	    i,
+#endif
 	    fdTypeStr[f->type],
-	    f->timeout_handler ? (int) (f->timeout - squid_curtime) / 60 : 0,
+	    f->timeout_handler ? (int) (f->timeout - squid_curtime) : 0,
 	    f->bytes_read,
 	    f->read_handler ? '*' : ' ',
 	    f->bytes_written,
@@ -459,6 +501,14 @@ info_get(StoreEntry * sentry)
 	runtime = 1.0;
     storeAppendPrintf(sentry, "Squid Object Cache: Version %s\n",
 	version_string);
+#ifdef _SQUID_WIN32_
+    if (WIN32_run_mode == _WIN_SQUID_RUN_MODE_SERVICE) {
+	storeAppendPrintf(sentry, "\nRunning as %s Windows System Service on %s\n",
+	    WIN32_Service_name, WIN32_OS_string);
+	storeAppendPrintf(sentry, "Service command line is: %s\n", WIN32_Service_Command_Line);
+    } else
+	storeAppendPrintf(sentry, "Running on %s\n", WIN32_OS_string);
+#endif
     storeAppendPrintf(sentry, "Start Time:\t%s\n",
 	mkrfc1123(squid_start.tv_sec));
     storeAppendPrintf(sentry, "Current Time:\t%s\n",
@@ -475,7 +525,13 @@ info_get(StoreEntry * sentry)
 	statCounter.icp.pkts_sent);
     storeAppendPrintf(sentry, "\tNumber of queued ICP replies:\t%u\n",
 	statCounter.icp.replies_queued);
-    storeAppendPrintf(sentry, "\tRequest failure ratio:\t%5.2f%%\n",
+#if USE_HTCP
+    storeAppendPrintf(sentry, "\tNumber of HTCP messages received:\t%u\n",
+	statCounter.htcp.pkts_recv);
+    storeAppendPrintf(sentry, "\tNumber of HTCP messages sent:\t%u\n",
+	statCounter.htcp.pkts_sent);
+#endif
+    storeAppendPrintf(sentry, "\tRequest failure ratio:\t%5.2f\n",
 	request_failure_ratio);
 
     storeAppendPrintf(sentry, "\tAverage HTTP requests per minute since start:\t%.1f\n",
@@ -542,6 +598,10 @@ info_get(StoreEntry * sentry)
 	statCPUUsage(5));
     storeAppendPrintf(sentry, "\tCPU Usage, 60 minute avg:\t%.2f%%\n",
 	statCPUUsage(60));
+#if HAVE_SBRK
+    storeAppendPrintf(sentry, "\tProcess Data Segment Size via sbrk(): %lu KB\n",
+	(unsigned long) (((char *) sbrk(0) - (char *) sbrk_start) >> 10));
+#endif
     storeAppendPrintf(sentry, "\tMaximum Resident Size: %d KB\n",
 	rusage_maxrss(&rusage));
     storeAppendPrintf(sentry, "\tPage faults with physical i/o: %d\n",
@@ -552,34 +612,34 @@ info_get(StoreEntry * sentry)
     storeAppendPrintf(sentry, "Memory usage for %s via mstats():\n",
 	appname);
     storeAppendPrintf(sentry, "\tTotal space in arena:  %6d KB\n",
-	ms.bytes_total >> 10);
+	(int) (ms.bytes_total >> 10));
     storeAppendPrintf(sentry, "\tTotal free:            %6d KB %d%%\n",
-	ms.bytes_free >> 10, percent(ms.bytes_free, ms.bytes_total));
+	(int) (ms.bytes_free >> 10), percent(ms.bytes_free, ms.bytes_total));
 #elif HAVE_MALLINFO && HAVE_STRUCT_MALLINFO
     mp = mallinfo();
     storeAppendPrintf(sentry, "Memory usage for %s via mallinfo():\n",
 	appname);
-    storeAppendPrintf(sentry, "\tTotal space in arena:  %6d KB\n",
-	mp.arena >> 10);
-    storeAppendPrintf(sentry, "\tOrdinary blocks:       %6d KB %6d blks\n",
-	mp.uordblks >> 10, mp.ordblks);
-    storeAppendPrintf(sentry, "\tSmall blocks:          %6d KB %6d blks\n",
-	mp.usmblks >> 10, mp.smblks);
-    storeAppendPrintf(sentry, "\tHolding blocks:        %6d KB %6d blks\n",
-	mp.hblkhd >> 10, mp.hblks);
-    storeAppendPrintf(sentry, "\tFree Small blocks:     %6d KB\n",
-	mp.fsmblks >> 10);
-    storeAppendPrintf(sentry, "\tFree Ordinary blocks:  %6d KB\n",
-	mp.fordblks >> 10);
-    t = mp.uordblks + mp.usmblks + mp.hblkhd;
+    storeAppendPrintf(sentry, "\tTotal space in arena:  %6ld KB\n",
+	(long int) (mp.arena >> 10));
+    storeAppendPrintf(sentry, "\tOrdinary blocks:       %6ld KB %6ld blks\n",
+	(long int) (mp.uordblks >> 10), (long int) mp.ordblks);
+    storeAppendPrintf(sentry, "\tSmall blocks:          %6ld KB %6ld blks\n",
+	(long int) (mp.usmblks >> 10), (long int) mp.smblks);
+    storeAppendPrintf(sentry, "\tHolding blocks:        %6ld KB %6ld blks\n",
+	(long int) (mp.hblkhd >> 10), (long int) mp.hblks);
+    storeAppendPrintf(sentry, "\tFree Small blocks:     %6ld KB\n",
+	(long int) (mp.fsmblks >> 10));
+    storeAppendPrintf(sentry, "\tFree Ordinary blocks:  %6ld KB\n",
+	(long int) (mp.fordblks >> 10));
+    t = (mp.uordblks + mp.usmblks + mp.hblkhd) >> 10;
     storeAppendPrintf(sentry, "\tTotal in use:          %6d KB %d%%\n",
-	t >> 10, percent(t, mp.arena + mp.hblkhd));
-    t = mp.fsmblks + mp.fordblks;
+	t, percent(t, (mp.arena + mp.hblkhd) >> 10));
+    t = (mp.fsmblks + mp.fordblks) >> 10;
     storeAppendPrintf(sentry, "\tTotal free:            %6d KB %d%%\n",
-	t >> 10, percent(t, mp.arena + mp.hblkhd));
-    t = mp.arena + mp.hblkhd;
+	t, percent(t, (mp.arena + mp.hblkhd) >> 10));
+    t = (mp.arena + mp.hblkhd) >> 10;
     storeAppendPrintf(sentry, "\tTotal size:            %6d KB\n",
-	t >> 10);
+	t);
 #if HAVE_EXT_MALLINFO
     storeAppendPrintf(sentry, "\tmax size of small blocks:\t%d\n", mp.mxfast);
     storeAppendPrintf(sentry, "\tnumber of small blocks in a holding block:\t%d\n",
@@ -595,10 +655,10 @@ info_get(StoreEntry * sentry)
 #endif /* HAVE_MALLINFO */
     storeAppendPrintf(sentry, "Memory accounted for:\n");
     storeAppendPrintf(sentry, "\tTotal accounted:       %6d KB\n",
-	statMemoryAccounted() >> 10);
-    storeAppendPrintf(sentry, "\tmemPoolAlloc calls: %d\n",
+	(int) (statMemoryAccounted() >> 10));
+    storeAppendPrintf(sentry, "\tmemPoolAlloc calls: %u\n",
 	mem_pool_alloc_calls);
-    storeAppendPrintf(sentry, "\tmemPoolFree calls: %d\n",
+    storeAppendPrintf(sentry, "\tmemPoolFree calls: %u\n",
 	mem_pool_free_calls);
 
     storeAppendPrintf(sentry, "File descriptor usage for %s:\n", appname);
@@ -616,6 +676,7 @@ info_get(StoreEntry * sentry)
 	RESERVED_FD);
     storeAppendPrintf(sentry, "\tStore Disk files open:                %4d\n",
 	store_open_disk_fd);
+    comm_select_status(sentry);
 
     storeAppendPrintf(sentry, "Internal Data Structures:\n");
     storeAppendPrintf(sentry, "\t%6d StoreEntries\n",
@@ -636,7 +697,7 @@ info_get(StoreEntry * sentry)
 #endif
 }
 
-#define XAVG(X) (dt ? (double) (f->X - l->X) / dt : 0.0)
+#define XAVG(X) (dt ? (f->X > l->X ? ((double) (f->X - l->X) / dt) : 0.0) : 0.0)
 static void
 statAvgDump(StoreEntry * sentry, int minutes, int hours)
 {
@@ -804,11 +865,10 @@ statAvgDump(StoreEntry * sentry, int minutes, int hours)
     storeAppendPrintf(sentry, "aborted_requests = %f/sec\n",
 	XAVG(aborted_requests));
 
-#if HAVE_POLL
-    storeAppendPrintf(sentry, "syscalls.polls = %f/sec\n", XAVG(syscalls.polls));
-#else
-    storeAppendPrintf(sentry, "syscalls.selects = %f/sec\n", XAVG(syscalls.selects));
-#endif
+    if (statCounter.syscalls.polls)
+	storeAppendPrintf(sentry, "syscalls.polls = %f/sec\n", XAVG(syscalls.polls));
+    if (statCounter.syscalls.selects)
+	storeAppendPrintf(sentry, "syscalls.selects = %f/sec\n", XAVG(syscalls.selects));
     storeAppendPrintf(sentry, "syscalls.disk.opens = %f/sec\n", XAVG(syscalls.disk.opens));
     storeAppendPrintf(sentry, "syscalls.disk.closes = %f/sec\n", XAVG(syscalls.disk.closes));
     storeAppendPrintf(sentry, "syscalls.disk.reads = %f/sec\n", XAVG(syscalls.disk.reads));
@@ -855,11 +915,15 @@ statInit(void)
     cachemgrRegister("vm_objects",
 	"In-Memory and In-Transit Objects",
 	stat_vmobjects_get, 0, 0);
-#if DEBUG_OPENFD
     cachemgrRegister("openfd_objects",
 	"Objects with Swapout files open",
 	statOpenfdObj, 0, 0);
-#endif
+    cachemgrRegister("pending_objects",
+	"Objects being retreived from the network",
+	statPendingObj, 0, 0);
+    cachemgrRegister("client_objects",
+	"Objects being sent to clients",
+	statClientsObj, 0, 0);
     cachemgrRegister("io",
 	"Server-side network read() size histograms",
 	stat_io_get, 0, 1);
@@ -946,6 +1010,8 @@ statAvgTick(void *notused)
 #elif HAVE_MALLINFO && HAVE_STRUCT_MALLINFO
 	struct mallinfo mp = mallinfo();
 	i = mp.arena;
+#elif HAVE_SBRK
+	i = (size_t) ((char *) sbrk(0) - (char *) sbrk_start);
 #endif
 	if (Config.warnings.high_memory < i)
 	    debug(18, 0) ("WARNING: Memory usage at %d MB\n", i >> 20);
@@ -1057,6 +1123,8 @@ statCountersHistograms(StoreEntry * sentry)
     statHistDump(&f->icp.reply_svc_time, sentry, NULL);
     storeAppendPrintf(sentry, "dns.svc_time histogram:\n");
     statHistDump(&f->dns.svc_time, sentry, NULL);
+    storeAppendPrintf(sentry, "select_fds_hist histogram:\n");
+    statHistDump(&f->select_fds_hist, sentry, NULL);
 }
 
 static void
@@ -1294,7 +1362,7 @@ statMedianSvc(int interval, int which)
 	x = statHistDeltaMedian(&l->dns.svc_time, &f->dns.svc_time);
 	break;
     default:
-	debug(49, 5) ("get_median_val: unknown type.\n");
+	debug(49, 5) ("statMedianSvc: unknown type.\n");
 	x = 0;
     }
     return x;
@@ -1385,7 +1453,7 @@ statByteHitRatio(int minutes)
      */
     cd = CountHist[0].cd.kbytes_recv.kb - CountHist[minutes].cd.kbytes_recv.kb;
     if (s < cd)
-	debug(18, 1) ("STRANGE: srv_kbytes=%d, cd_kbytes=%d\n", s, cd);
+	debug(18, 1) ("STRANGE: srv_kbytes=%d, cd_kbytes=%d\n", (int) s, (int) cd);
     s -= cd;
 #endif
     if (c > s)
@@ -1403,13 +1471,14 @@ statClientRequests(StoreEntry * s)
     StoreEntry *e;
     int fd;
     for (i = ClientActiveRequests.head; i; i = i->next) {
+	const char *p = NULL;
 	http = i->data;
 	assert(http);
 	conn = http->conn;
 	storeAppendPrintf(s, "Connection: %p\n", conn);
 	if (conn) {
 	    fd = conn->fd;
-	    storeAppendPrintf(s, "\tFD %d, read %d, wrote %d\n", fd,
+	    storeAppendPrintf(s, "\tFD %d, read %" PRINTF_OFF_T ", wrote %" PRINTF_OFF_T "\n", fd,
 		fd_table[fd].bytes_read, fd_table[fd].bytes_written);
 	    storeAppendPrintf(s, "\tFD desc: %s\n", fd_table[fd].desc);
 	    storeAppendPrintf(s, "\tin: buf %p, offset %ld, size %ld\n",
@@ -1438,6 +1507,23 @@ statClientRequests(StoreEntry * s)
 	    (long int) http->start.tv_sec,
 	    (int) http->start.tv_usec,
 	    tvSubDsec(http->start, current_time));
+	if (http->request->auth_user_request)
+	    p = authenticateUserRequestUsername(http->request->auth_user_request);
+	else if (http->request->extacl_user) {
+	    p = http->request->extacl_user;
+	}
+	if (!p && conn->rfc931[0])
+	    p = conn->rfc931;
+#if USE_SSL
+	if (!p)
+	    p = sslGetUserEmail(fd_table[conn->fd].ssl);
+#endif
+	if (!p)
+	    p = dash_str;
+	storeAppendPrintf(s, "username %s\n", p);
+#if DELAY_POOLS
+	storeAppendPrintf(s, "delay_pool %d\n", delayClient(http) >> 16);
+#endif
 	storeAppendPrintf(s, "\n");
     }
 }
@@ -1524,10 +1610,10 @@ statGraphDump(StoreEntry * e)
     GENGRAPH(cputime, "cputime", "CPU utilisation");
 }
 
-int
+#endif /* STAT_GRAPHS */
+
+size_t
 statMemoryAccounted(void)
 {
-    memTotalAllocated();
+    return memTotalAllocated();
 }
-
-#endif /* STAT_GRAPHS */
