@@ -1,6 +1,6 @@
 
 /*
- * $Id: typedefs.h,v 1.147 2006/06/04 02:50:05 hno Exp $
+ * $Id: typedefs.h,v 1.157 2007/12/14 20:05:24 hno Exp $
  *
  *
  * SQUID Web Proxy Cache          http://www.squid-cache.org/
@@ -177,6 +177,7 @@ typedef struct _request_t request_t;
 typedef struct _AccessLogEntry AccessLogEntry;
 typedef struct _cachemgr_passwd cachemgr_passwd;
 typedef struct _refresh_t refresh_t;
+typedef struct _refresh_cc refresh_cc;
 typedef struct _CommWriteStateData CommWriteStateData;
 typedef struct _ErrorState ErrorState;
 typedef struct _dlink_node dlink_node;
@@ -213,6 +214,7 @@ typedef struct _link_list link_list;
 typedef struct _storefs_entry storefs_entry_t;
 typedef struct _storerepl_entry storerepl_entry_t;
 typedef struct _diskd_queue diskd_queue;
+typedef struct _logfile_buffer logfile_buffer_t;
 typedef struct _Logfile Logfile;
 typedef struct _logformat_token logformat_token;
 typedef struct _logformat logformat;
@@ -224,6 +226,7 @@ typedef struct _RemovalPolicyNode RemovalPolicyNode;
 typedef struct _RemovalPolicySettings RemovalPolicySettings;
 typedef struct _errormap errormap;
 typedef struct _PeerMonitor PeerMonitor;
+typedef struct _vary_id_t vary_id_t;
 
 typedef struct _http_version_t http_version_t;
 
@@ -276,6 +279,7 @@ typedef void STRCB(void *their_data, const char *buf, ssize_t len);
 typedef void SIH(storeIOState *, void *);	/* swap in */
 typedef int QS(const void *, const void *);	/* qsort */
 typedef void STCB(void *, char *, ssize_t);	/* store callback */
+typedef void STHCB(void *, HttpReply *);	/* store callback */
 typedef void STABH(void *);
 typedef void ERCB(int fd, void *, size_t);
 typedef void OBJH(StoreEntry *);
@@ -289,6 +293,7 @@ typedef void HLPCMDOPTS(int *argc, char **argv);
 typedef void IDNSCB(void *, rfc1035_rr *, int, const char *);
 
 typedef void STINIT(SwapDir *);
+typedef void STCHECKCONFIG(SwapDir *);
 typedef void STNEWFS(SwapDir *);
 typedef void STDUMP(StoreEntry *, SwapDir *);
 typedef void STFREE(SwapDir *);
@@ -310,6 +315,7 @@ typedef void STOBJCLOSE(SwapDir *, storeIOState *);
 typedef void STOBJREAD(SwapDir *, storeIOState *, char *, size_t, squid_off_t, STRCB *, void *);
 typedef void STOBJWRITE(SwapDir *, storeIOState *, char *, size_t, squid_off_t, FREE *);
 typedef void STOBJUNLINK(SwapDir *, StoreEntry *);
+typedef void STOBJRECYCLE(SwapDir *, StoreEntry *);
 
 typedef void STLOGOPEN(SwapDir *);
 typedef void STLOGCLOSE(SwapDir *);
@@ -345,6 +351,7 @@ typedef void AUTHSFREECONFIG(authScheme *);
 typedef char *AUTHSUSERNAME(auth_user_t *);
 typedef void AUTHSONCLOSEC(ConnStateData *);
 typedef void AUTHSPARSE(authScheme *, int, char *);
+typedef void AUTHSCHECKCONFIG(authScheme *);
 typedef void AUTHSINIT(authScheme *);
 typedef void AUTHSREQFREE(auth_user_request_t *);
 typedef void AUTHSSETUP(authscheme_entry_t *);
@@ -402,4 +409,17 @@ typedef void ERRMAPCB(StoreEntry *, int body_offset, squid_off_t content_length,
 typedef struct _VaryData VaryData;
 typedef void STLVCB(VaryData * vary, void *cbdata);
 
+typedef struct _HttpMsgBuf HttpMsgBuf;
+
+typedef struct rms rms_t;
+
+typedef void LOGLINESTART(Logfile *);
+typedef void LOGWRITE(Logfile *, const char *, size_t len);
+typedef void LOGLINEEND(Logfile *);
+typedef void LOGFLUSH(Logfile *);
+typedef void LOGROTATE(Logfile *);
+typedef void LOGCLOSE(Logfile *);
+
+typedef void REFRESHCHECK(void *data, int fresh, const char *log);
+typedef struct _refresh_check_helper refresh_check_helper;
 #endif /* SQUID_TYPEDEFS_H */
